!---------------------------------------------------------------------------------------
! FLEXINVERT : background
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculates the background mixing ratio for each trajectory at each receptor by
!! multiplying the sensitivities to the initial mixing ratios with the mixing ratios from
!! another model
!!
!! Calls the following subroutines:
!!  - readinit: reads the gridded FLEXPART output files grid_initial_001 containing the 
!!    sensitivity to the inital mixing ratios for a given receptor point
!!  - initial: reads mixing ratios from a global model and converts these to the flexpart
!!    grid for use with the initial sensitivities
!!    (note: to change to a custom initial mixing ratio file modify the subroutine 
!!     e.g. readlmdz.f90 in initial.f90)
!---------------------------------------------------------------------------------------

subroutine background(bkgnd)

  use mod_var
  use mod_dates

  implicit none

  character(len=100) :: dirflexrec,filename,datestr
  integer :: numpoint,eomday
  integer :: yyyymmddhh,yyyymm,yyyy,mm,dd,hh,nr,nt,i,it,ix,jy,kz,ind,indo
  real(kind=8) :: jd,jtime
  real, dimension(nzmax) :: outheight
  integer, dimension(maxpoint) :: releases
  real, dimension(maxrows,ncrs) :: bkgnd
  real, dimension(numxgrid*numygrid) :: bkgndvec
  real, dimension(:,:,:,:), allocatable :: init
  integer, parameter :: ntmax=1000  ! max no. timesteps in initial mixing ratio file

  write(logid,*) 'Calculating background from initial conditions...'

  allocate(indexbkg(maxrows,2))

  ! loop over receptors/stations

  ind=0
  do nr=1,numrec
    jd=juldatei
    indo=0

    ! loop over months in inversion

    do while(jd.lt.juldatef)
      yyyymmddhh=caldate(jd)
      yyyymm=yyyymmddhh/10000
      yyyy=yyyymmddhh/1000000
      mm=(yyyymmddhh-1000000*yyyy)/10000
      dd=(yyyymmddhh-1000000*yyyy-10000*mm)/100
      hh=(yyyymmddhh-yyyy*1000000-mm*10000-dd*100)
      write(datestr,'(i6)') yyyymm
      eomday=calceomday(yyyymm)
      dirflexrec=trim(dirflex)//trim(recname(nr))//'/'//trim(datestr)//'/'

      filename=trim(dirflexrec)//'header'
      call readheader(filename,numpoint,outheight,releases)

      ! read grid_init files
      allocate(init(numxgrid,numygrid,numzgrid,numpoint))
      call readinit(dirflexrec,numpoint,init)

      ! read initial mixing ratios and converts to flexpart grid  
      call initconc(outheight,ntmax,numpoint,jd)

      ! convert from ppt to output unit for multiplication with immr 
      init=init*coef

      ! calculate background contribution to mixing ratio

      ! loop over trajectories for the current month
      do nt=1,numpoint
        ind=ind+1
        indo=indo+1
        bkgndvec(:)=0.
        ! calculate time index to initial mixing ratio
        ! subtract length of back trajectory
        ! note: releases gives hour of trajectory since 1st release in backwards time
        jtime=float(yyyy)+float(mm-1)/12.+float(dd-1)/365.+float(hh)/(365.*24.)+&
              &float(eomday*24+releases(nt)/3600)/24./365.-(trajlen/365.)
        it=minloc(abs(timini-jtime),dim=1,mask=abs(timini-jtime).eq.minval(abs(timini-jtime)))
        do kz=1,numzgrid
          do jy=1,numygrid
            do ix=1,numxgrid
              if(isnan(immr(ix,jy,kz,it))) go to 10 
              bkgndvec((jy-1)*numxgrid+ix)=bkgndvec((jy-1)*numxgrid+ix)+init(ix,jy,kz,nt)*immr(ix,jy,kz,it)
10            continue
            end do
          end do
        end do
        ! record index of background
        indexbkg(ind,1)=nr
        if(trim(rectyp(nr)).eq.'FM'.or.trim(rectyp(nr)).eq.'CM') then
          ! index by model time step in hours
          indexbkg(ind,2)=int((jd-juldatei)*24+nt*relperiod)
        else
          ! index by number of observations 
          ! note: assumes one back trajectory for each observation
          indexbkg(ind,2)=indo
        endif
        ! convert background vector to coarse resolution
        bkgnd(ind,:)=matmul(real(gridcrs),bkgndvec)
      end do ! numpoint
  
      jd=jd+eomday

      deallocate(init)
      deallocate(immr)
      deallocate(timini)

    end do ! jd
  end do  ! nr

  open(600,file=trim(dirinvert)//'indexbkg.txt',status='replace',action='write')
  do i=1,ind
    write(600,fmt='(I2,1X,I5)') indexbkg(i,1),indexbkg(i,2)
  end do
  close(600)


end subroutine background  

