!---------------------------------------------------------------------------------------
! FLEXINVERT : bayesian1
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Finds the solution vector x corresponding to the probability maximum, that is
!! the solution which minimizes the cost function:
!!
!! J(x) = (x-xb)^T*B^(-1)*(x-xb) + (Hx-y)^T*R^(-1)*(Hx-y)
!!
!! where:
!!   xb = prior/background value of x (nn)
!!   B  = prior error covariance matrix (nn,nn)
!!   H  = kernel matrix (nm,nn)
!!   R  = observation error covariance matrix (nm,nm)
!!   y  = observation vector (nm)
!!
!! Optimal value of x:
!!
!!   x = xb + BH^T(HBH^T + R)^(-1)(y - Hxb)
!!
!! and its covariance:
!!
!!   cova = B - BH^T(HBH^T + R)^(-1)HB
!!
!! Ref: Tarantola, Inverse Problem Theory, 2005
!---------------------------------------------------------------------------------------

subroutine bayesian1(nn,np,nm,covr,yobs,kern,xb,x,cova,gain)

  use mod_var

  implicit none

  integer :: i,j,k,l,jt,it,nn,np,nm,N,lda,info
  real, dimension(nn,nn) :: cova
  real, dimension(nm,nm) :: covr
  real, dimension(np) :: yobs
  real, dimension(nm) :: ymod,yerr
  real, dimension(np,nn) :: kern
  real, dimension(nn,nm) :: kernt
  real, dimension(nn) :: x,xb
  real(kind=8), dimension(1,neig) :: tmp
  real(kind=8), dimension(1,nbox) :: tmp1
  real(kind=8), dimension(nm,nn) :: kernb
  real(kind=8), dimension(nn,nm) :: bkernt,gain
  real(kind=8), dimension(nm,nm) :: M,mkern,imkern

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  character(len=100) :: fileout,varname,varunit,comment
  character(len=20) :: dname1,dname2  
  real(kind=8), dimension(nbox*nemsteps,nbox*nemsteps) :: covb 
  integer, dimension(nbox*nemsteps) :: dimboxtstep
  real(kind=8), dimension(nm,nbox*nemsteps) :: kernb_test
  real(kind=8), dimension(nbox*nemsteps,nm) :: bkernt_test
  !--------------------------------------------------------------

  write(logid,*) 'Performing the inversion step...'

  ! calculate y - Hxb
  ymod(:)=0.
  yerr(:)=0.
  do i=1,nm
    ymod(i)=dot_product(kern(i,:),xb)
    yerr(i)=yobs(i)-ymod(i)
  end do
  
  ! calculate H^T
  kernt=transpose(kern(1:nm,:))

  ! calculate BH^T
  ! use eigendecomposition of B
  bkernt(:,:)=0.
  do jt=1,nemsteps
    do i=1,nbox
      ! row vector of U*D
      do j=1,neig
        tmp(1,j)=bevec(i,j)*beval(j)
      end do
      ! row vector of U*D*U^T
      do k=1,nbox
        tmp1(1:1,k:k)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
      end do
      do it=1,nemsteps
        do l=1,nm
          bkernt((jt-1)*nbox+i,l:l)=bkernt((jt-1)*nbox+i,l:l)+cortime(it,jt)*matmul(tmp1,kernt((it-1)*nbox+1:it*nbox,l))
        end do
      end do ! it
    end do ! i
  end do ! jt
  ! finish for nbkgnd assuming 1% errors
  do i=1,nbkgnd
    bkernt(nemsteps*nbox+i,:)=kernt(nemsteps*nbox+i,:)*0.01**2
  end do

  ! calculate (HBH^T + R)
  mkern=matmul(kern(1:nm,:),bkernt)+dble(covr)

  ! calculate condition number of (HBH^T + R)
  call calcconditnum(nm,mkern)

  ! calculate (HBH^T + R)^-1
  ! use Cholesky factorization
  N=nm            ! order of (HBH^T + R)
  M=mkern         ! symmetric matrix
  lda=nm          ! leading dimension of (HBH^T + R)
  call dpotrf('L',N,M,lda,info)
  if(info.gt.0) then 
    write(logid,*) 'Error in calculating lower Cholesky triangle'
    stop
  endif
  call dpotri('L',N,M,lda,info)
  if(info.gt.0) then 
    write(logid,*) 'Error in inverting (HBH^T + R)'
    stop
  endif
  ! returns lower ('L') triangle of inverse of M
  imkern=M
  do i=1,nm
    do j=1,i
      imkern(j,i)=M(i,j)
    end do
  end do

  ! calculate Gain matrix BH^T(HBH^T + R)^(-1)
  gain=matmul(bkernt,imkern)

  ! calculate x = xb + BH^T(HBH^T + R)^(-1)(y-Hxb)
  x(:)=0.
  do i=1,nn
    x(i)=xb(i)+dot_product(real(gain(i,:),kind=4),yerr)
  end do

  ! calculate HB
  ! use eigendecomposition of B
  kernb(:,:)=0.
  do jt=1,nemsteps
    do k=1,nbox
      ! calculate column vector of U*D*U^T
      do i=1,nbox
        ! calculate row vector of U*D
        do j=1,neig
          tmp(1,j)=bevec(i,j)*beval(j)
        end do
        tmp1(1:1,i:i)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
      end do
      do it=1,nemsteps
        do l=1,nm
          kernb(l:l,(jt-1)*nbox+k)=kernb(l:l,(jt-1)*nbox+k)+&
                &cortime(it,jt)*matmul(real(kern(l,((it-1)*nbox+1):(it*nbox)),kind=8),transpose(tmp1))
        end do
      end do
    end do ! k
  end do ! jt
  ! finish for nbkgnd assuming 1% errors
  do i=1,nbkgnd
    kernb(:,nemsteps*nbox+i)=real(kern(1:nm,nemsteps*nbox+i)*0.01**2,kind=8)
  end do

  ! calculate B - BH^T(HBH^T + R)^(-1)HB
  ! use eigendecomposition of B
  cova(:,:)=0.

  ! TESTING
  covb(:,:)=0.
  ! 

  do jt=1,nemsteps
    do k=1,nbox
      do i=1,nbox
        ! calculate row vector of U*D
        do j=1,neig
          tmp(1,j)=bevec(i,j)*beval(j)
        end do
        tmp1(1:1,i:i)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
      end do
      do it=1,nemsteps

        ! TESTING
        covb((jt-1)*nbox+1:(jt*nbox),(it-1)*nbox+k)=cortime(it,jt)*tmp1(1,:)
        !

        cova(((jt-1)*nbox+1):(jt*nbox),(it-1)*nbox+k)=cortime(it,jt)*tmp1(1,:)-&
             &matmul(gain((it-1)*nbox+k,:),kernb(:,(jt-1)*nbox+1:jt*nbox))

      end do
    end do ! k
  end do ! jt

  ! finish for nbkgnd assuming 1% errors
  do i=nemsteps*nbox+1,nemsteps*nbox+nbkgnd
    do j=1,nemsteps*nbox+nbkgnd
      if(i==j) then
        cova(i:i,j)=0.01**2-matmul(gain(i,:),kernb(:,j:j))
      else
        cova(i:i,j)=0.-matmul(gain(i,:),kernb(:,j:j))
        ! since cova must be symmetric
        cova(j,i:i)=0.-matmul(gain(i,:),kernb(:,j:j))
      endif
    end do
  end do

  write(logid,*) 'Congratulations you have optimized the control vector'

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  ! B (covb) 
  fileout='covb_eigen.nc'
  varname='covb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='covb reconstructed from eigendecomposition'
  do i=1,nbox*nemsteps
    dimboxtstep(i)=i
  end do
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(covb,4),&
                   &nbox*nemsteps,nbox*nemsteps,dimboxtstep,dimboxtstep,dname1,dname2,comment)
  ! HB (kernb from direct multiplication)
  kernb_test=matmul(kern(1:nm,1:nbox*nemsteps),covb)
  fileout='hb_test.nc'
  varname='hb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='HB from direct multiplication'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(kernb_test,4),&
                   &nm,nbox*nemsteps,dimr,dimboxtstep,dname1,dname2,comment)  
  ! HB (kernb from eigen calculations)
  fileout='hb_eigen.nc'
  varname='hb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='HB reconstructed from eigendecomposition'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(kernb(:,1:nbox*nemsteps),4),&
                   &nm,nbox*nemsteps,dimr,dimboxtstep,dname1,dname2,comment)  
  ! BH^T (bkernt from direct multiplication)
  bkernt_test=matmul(covb,kernt(1:nbox*nemsteps,1:nm))
  fileout='bht_test.nc'
  varname='bht'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='BH^T from direct multiplication'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(bkernt_test,4),&
                   &nbox*nemsteps,nm,dimboxtstep,dimr,dname1,dname2,comment)
  ! BH^T (bkernt from eigen calculations)
  fileout='bht_eigen.nc'
  varname='bht'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='BH^T reconstructed from eigendecomposition'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(bkernt,4),&
                   &nbox*nemsteps,nm,dimboxtstep,dimr,dname1,dname2,comment)
  !--------------------------------------------------------------

end subroutine bayesian1

