!---------------------------------------------------------------------------------------
! FLEXINVERT : bayesian2
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Finds the solution vector x corresponding to the probability maximum, that is
!! the solution which minimizes the cost function:
!!
!! J(x) = (x-xb)^T*B^(-1)*(x-xb) + (Hx-y)^T*R^(-1)*(Hx-y)
!!
!! where:
!!   xb = prior/background value of x (nn)
!!   B  = prior error covariance matrix (nn,nn)
!!   H  = kernel matrix (nm,nn)
!!   R  = observation error covariance matrix (nm,nm)
!!   y  = observation vector (nm)
!!
!! Optimal value of x:
!! only inverting symmetric matrices B, R and (H^TR^(-1)H + B^(-1)) 
!!
!!   x = xb + (H^TR^(-1)H + B^(-1))^(-1)H^TR^(-1)(y-Hxb)
!!
!! and its covariance:
!!
!!   cova = (H^TR^(-1)H + B^(-1))^(-1)
!!
!! Ref: Tarantola, Inverse Problem Theory, 2005
!---------------------------------------------------------------------------------------

subroutine bayesian2(nn,np,nm,covr,yobs,kern,xb,x,cova,gain)

  use mod_var

  implicit none

  integer :: i,j,k,it,jt,nn,np,nm,M,N,lda,info
  real, dimension(nn,nn) :: cova
  real(kind=8), dimension(nn,nn) :: akern,A,iakern
  real, dimension(nm,nm) :: covr
  real(kind=8), dimension(nm,nm) :: R,icovr
  real, dimension(np) :: yobs
  real, dimension(nm) :: ymod,yerr
  real, dimension(np,nn) :: kern
  real, dimension(nn,nm) :: kernt
  real(kind=8), dimension(nn,nm) :: lhs,gain
  real, dimension(nn) :: x,xb
  real(kind=8), dimension(1,neig) :: tmp
  real(kind=8), dimension(1,nbox) :: tmp1
  integer :: ldu,ldvt,lwork
  integer, dimension(:), allocatable :: iwork
  real(kind=8), dimension(nn) :: sigma
  real(kind=8), dimension(:), allocatable :: work
  real(kind=8), dimension(:,:), allocatable :: U,UT,V,VT,iS,ViS

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  character(len=100) :: fileout,varname,varunit,comment 
  character(len=20) :: dname1,dname2  
  integer, dimension(nbox*nemsteps) :: dimboxtstep
  real(kind=8), dimension(nbox*nemsteps,nbox*nemsteps) :: icovb
  real(kind=8), dimension(nbox*nemsteps,nbox*nemsteps) :: akern_test
  !--------------------------------------------------------------

  write(logid,*) 'Performing the inversion step...'

  ! calculate y - Hxb
  ymod(:)=0.
  yerr(:)=0.
  do i=1,nm
    ymod(i)=dot_product(kern(i,:),xb)
    yerr(i)=yobs(i)-ymod(i)
  end do
  
  ! calculate H^T
  kernt=transpose(kern(1:nm,:))

  ! calculate R^(-1)
  ! use Cholesky factorization 
  N=nm           ! order of R
  R=dble(covr)   ! symmetric matrix
  lda=nm         ! leading dimension of R
  call dpotrf('L',N,R,lda,info)
  call dpotri('L',N,R,lda,info)
  if(info.gt.0) then 
    write(logid,*) 'Error in inverting R'
    stop
  endif
  ! returns lower ('L') triangle of inverse of R
  icovr=R
  do i=1,nm
    do j=1,i 
      icovr(j,i)=R(i,j)
    end do
  end do

  ! calculate (H^TR^(-1)H + B^(-1))
  ! use eigendecomposition of B

  ! TESTING 
  icovb(:,:)=0.
  !

  lhs=matmul(dble(kernt),icovr)
  do jt=1,nemsteps
    do k=1,nbox
      ! calculate column vector of U*D^(-1)*U^T
      do i=1,nbox
        do j=1,neig
          ! calculate row vector of U*D^(-1)
          tmp(1,j)=bevec(i,j)*(1./beval(j))
        end do
        tmp1(1:1,i:i)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
      end do
      do it=1,nemsteps

        ! TESTING
        icovb((jt-1)*nbox+1:(jt*nbox),(it-1)*nbox+k)=icortime(it,jt)*tmp1(1,:)
        !

        akern(((jt-1)*nbox+1):(jt*nbox),(it-1)*nbox+k)=icortime(it,jt)*tmp1(1,:)+&
             &matmul(lhs((it-1)*nbox+k,:),real(kern(1:nm,(it-1)*nbox+1:it*nbox),kind=8))
      end do
    end do ! k
  end do ! jt

  ! finish for nbkgnd assuming 1% errors
  do i=1,nbkgnd
    do j=1,nemsteps*nbox+nbkgnd
      if(i.ne.j) then
        akern(nemsteps*nbox+i,j:j)=matmul(lhs(nemsteps*nbox+i,:),real(kern(1:nm,j:j),kind=8))
        ! H^TR^(-1)H is symmetric
        akern(j:j,nemsteps*nbox+i)=matmul(lhs(nemsteps*nbox+i,:),real(kern(1:nm,j:j),kind=8))
      else
        akern(nemsteps*nbox+i,j:j)=matmul(lhs(nemsteps*nbox+i,:),real(kern(1:nm,j:j),kind=8))+1./0.01**2
      endif
    end do
  end do

  ! calculate condition number of (H^TR^(-1)H + B^(-1))
  call calcconditnum(nn,akern)

  ! calculate inverse of (H^TR^(-1)H + B^(-1))
  ! use Cholesky factorization
  N=nn           ! order of (H^TR^(-1)H + B^(-1))
  A=akern        ! symmetric matrix
  lda=nn         ! leading dimension of (H^TR^(-1)H + B^(-1))
  call dpotrf('L',N,A,lda,info)
  if(info.gt.0) then
    write(logid,*) 'Error in calculating lower Cholesky triangle'
    stop
  endif
  call dpotri('L',N,A,lda,info)
  if(info.gt.0) then
    write(logid,*) 'Error in inverting (H^TR^(-1)H + B^(-1))'
    stop
  endif
  ! returns lower ('L') triangle of inverse of A
  iakern=A
  do i=1,nn
    do j=1,i
      iakern(j,i)=A(i,j)
    end do
  end do

  ! calculate inverse of (H^TR^(-1)H + B^(-1))
  ! use SVD (Cholesky fails for non-positive-definite matrices)
!  M=nn           ! number of rows of A
!  N=nn           ! number of columns of A
!  A=akern        ! matrix of dimension M x N
!  lda=M          ! leading dimension of A
!  ldu=M          ! leading dimension of U
!  ldvt=N         ! leading dimension of VT
!  lwork=3*min(M,N)+max(max(M,N),4*min(M,N)*min(M,N)+4*min(M,N))
!  allocate(work(lwork))
!  allocate(iwork(8*min(M,N)))
!  allocate(U(M,M))
!  allocate(VT(N,N))
!  call dgesdd('A',M,N,A,lda,sigma,U,ldu,VT,ldvt,work,lwork,iwork,info)
!  if(info.ne.0) then
!    print*, 'Error in singular value decomposition'
!    if(info.lt.0) print*, 'the argument in position ',info,' had an illegal value'
!    if(info.gt.0) print*, 'the algorithm failed to converge'
!    stop
!  endif
!  allocate(UT(M,M))
!  allocate(V(N,N))
!  UT=transpose(U)
!  V=transpose(VT)
!  deallocate(U)
!  deallocate(VT)
!  allocate(iS(N,N))
!  iS(:,:)=0.
!  do i=1,nn
!    if(sigma(i).ne.0.) iS(i,i)=1./sigma(i)
!  end do
!  allocate(ViS(N,N))
!  ViS=matmul(V,iS)
!  iakern=matmul(ViS,UT)
!  deallocate(UT)
!  deallocate(V)
!  deallocate(iS)
!  deallocate(ViS)
!  deallocate(iwork)
!  deallocate(work)

  ! calculate Gain matrix (H^TR^(-1)H + B^(-1))^(-1)H^TR^(-1)
  gain=matmul(iakern,lhs)

  ! calculate x = xb + G(y - Hxb)
  x(:)=0.
  do i=1,nn
    x(i)=xb(i)+dot_product(real(gain(i,:),kind=4),yerr)
  end do

  ! posterior covariance matrix
  cova=real(iakern,kind=4)

  write(logid,*) 'Congratulations you have optimized the control vector'

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  ! B^(-1) (icovb)
  fileout='icovb_eigen.nc'
  varname='icovb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='icovb reconstructed from eigendecomposition'
  do i=1,nbox*nemsteps
    dimboxtstep(i)=i
  end do
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(icovb,4),&
                   &nbox*nemsteps,nbox*nemsteps,dimboxtstep,dimboxtstep,dname1,dname2,comment)  
  ! H^TR^(-1)H+B^(-1) (akern from direct calculation)  
  akern_test=matmul(lhs(1:nbox*nemsteps,:),kern(1:nm,1:nbox*nemsteps))+icovb
  fileout='akern_test.nc'
  varname='akern'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='akern from direct calculation'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(akern_test,4),&
                   &nbox*nemsteps,nbox*nemsteps,dimboxtstep,dimboxtstep,dname1,dname2,comment)
  ! H^TR^(-1)H+B^(-1) (akern from eigen calculations)
  fileout='akern_eigen.nc'
  varname='akern'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='akern reconstructed from eigendecomposition'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(akern,4),&
                   &nbox*nemsteps,nbox*nemsteps,dimboxtstep,dimboxtstep,dname1,dname2,comment)
  !--------------------------------------------------------------

end subroutine bayesian2
