!---------------------------------------------------------------------------------------
! FLEXINVERT : calccost
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculates the cost of the Bayesian function:
!! J(x) = (x-xb)^T*B^(-1)*(x-xb) + (Hx-y)^T*R^(-1)*(Hx-y)
!!
!! Options:
!!   stat: status 1 = a priori, 2 = a posteriori
!---------------------------------------------------------------------------------------

subroutine calccost(stat,nn,np,nm,x,xb,yobs,kern,covr,cost)

  use mod_var

  implicit none

  integer :: i,j,k,it,jt,stat,nn,np,nm,M,N,lda,info
  real, dimension(nn) :: x,xb
  real, dimension(1,nn) :: xerr
  real, dimension(1,nn) :: lhsb
  real, dimension(np) :: yobs
  real, dimension(nm) :: yerr,lhsr
  real(kind=8), dimension(1,neig) :: tmp
  real(kind=8), dimension(nbox,1) :: tmp1
  real, dimension(np,nn) :: kern
  real, dimension(nm,nm) :: covr
  real(kind=8), dimension(nm,nm) :: R,icovr
  real, dimension(1) :: costr,cost
  real, dimension(1,1) :: costb

  ! Calculate R^(-1)
  ! use Cholesky factorization 
  N=nm         ! order of R
  R=dble(covr) ! symmetric matrix
  lda=nm       ! leading dimension of R
  call dpotrf('L',N,R,lda,info)
  call dpotri('L',N,R,lda,info)
  if(info.gt.0) write(logid,*) 'Error in inverting R'
  ! returns lower ('L') triangle of inverse of R
  icovr=R
  do i=1,nm
    do j=1,i
      icovr(j,i)=R(i,j)
    end do
  end do

  ! Cost a priori
  ! J = (Hxb-y)^T*R^(-1)*(Hxb-y)
  if(stat.eq.1) then
    yerr=matmul(kern(1:nm,:),xb)-yobs(1:nm)
    lhsr=matmul(yerr,real(icovr,kind=4))
    cost=matmul(lhsr,reshape(yerr,(/nm,1/)))
    write(logid,fmt='(1X,A20,F11.3)') 'Cost a priori (Jy): ',cost
  endif

  ! Cost a posteriori
  ! J = (x-xb)^T*B^(-1)*(x-xb) + (Hx-y)^T*R^(-1)*(Hx-y)
  if(stat.eq.2) then
    yerr=matmul(kern(1:nm,:),x)-yobs(1:nm)
    lhsr=matmul(yerr,real(icovr,kind=4))
    costr=matmul(lhsr,reshape(yerr,(/nm,1/)))
    xerr(1,:)=x-xb
    ! calculate (x-xb)*B^(-1) using eigendecomposition of B
    lhsb(:,:)=0.
    do jt=1,nemsteps
      do k=1,nbox 
        ! calculate column vector of U*D^(-1)*U^T
        do i=1,nbox
          ! calculate row vector of U*D^(-1)
          do j=1,neig
            tmp(1:1,j:j)=bevec(i,j)*(1./beval(j))
          end do
          tmp1(i:i,1:1)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
        end do
        do it=1,nemsteps
          lhsb(1:1,(jt-1)*nbox+k:(jt-1)*nbox+k)=lhsb(1:1,(jt-1)*nbox+k:(jt-1)*nbox+k)+&
               &icortime(it,jt)*matmul(xerr(1:1,(it-1)*nbox+1:it*nbox),real(tmp1,kind=4))
        end do
      end do ! k 
    end do ! jt
    ! finish for nbkgnd assuming 1% errors
    do i=1,nbkgnd
      lhsb(1:1,nemsteps*nbox+i)=xerr(1,nemsteps*nbox+i)/(0.01**2)
    end do
    costb(1:1,1:1)=matmul(lhsb,reshape(xerr,(/nn,1/)))
    cost=costb(1,1)+costr
    write(logid,fmt='(1X,A24,F11.3)') 'Cost a posteriori (Jy): ',costr
    write(logid,fmt='(1X,A24,F11.3)') 'Cost a posteriori (Jx): ',costb
    write(logid,fmt='(1X,A29,F11.3)') 'Cost a posteriori total (J): ',cost
    write(logid,fmt='(1X,A36,F11.3)') 'Degrees of freedom of signal (DFS): ',costb
  endif

end subroutine calccost



