!---------------------------------------------------------------------------------------
! FLEXINVERT : calcperturb
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Adds random perturbation to the control vector (prior emissions) and observations
!! with the same error covariance statistics as the B and R matrices, respectively
!!
!! xb' = xb + B^0.5*rb
!!
!! y' = y + R^0.5*rr
!!
!! where rb and rr are random vectors of same length as xb and y, respectively
!---------------------------------------------------------------------------------------

subroutine calcperturb(nn,np,nm,covr,xb,yobs,ymod)

  use mod_var

  implicit none

  integer :: i,j,k,it,jt,ind,nn,np,nm,ierr
  integer :: N,M,lda,ldz,il,iu,lwork,liwork,info
  character(len=5) :: seedstr
  real, dimension(nn) :: xb,rbvec
  real, dimension(np) :: yobs,ymod
  real, dimension(nm) :: rrvec
  real(kind=8), dimension(nn,nn) :: A
  real(kind=8) :: vl,vu,abstol
  real, dimension(nm,nm) :: covr
  real(kind=8), dimension(nm,nm) :: sqrtcovr
  real(kind=8),dimension(nm,nm) :: U
  real(kind=8), dimension(1,neig) :: tmp
  real(kind=8), dimension(1,nbox) :: tmp1
  real :: ri
  integer, dimension(:), allocatable :: isuppz
  real(kind=8), dimension(:), allocatable :: work
  real(kind=8), dimension(:,:), allocatable :: sqrtw
  real(kind=8), dimension(:), allocatable :: W
  real(kind=8), dimension(:,:), allocatable :: Z
  integer, dimension(:), allocatable :: iwork
  integer :: ssize
  integer, dimension(:), allocatable :: seedvec

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  character(len=100) :: fileout,varname,varunit,comment  
  character(len=20) :: dname1,dname2  
  real(kind=8), dimension(nbox,nbox) :: sqrtb  
  !--------------------------------------------------------------


  write(logid,*) 'Adding random perturbation to xb and yobs'

  ! Record true emissions before perturbation

  open(100,file=trim(dirinvert)//'true.txt',status='replace',action='write',iostat=ierr)
  do i=1,nbox*nemsteps
    write(100,*) xb(i)/numscale
  end do
  do i=1,nbkgnd
    write(100,*) xb(nbox*nemsteps+i)
  end do
  close(100)

  ! for synthetic data test set yobs equal to ymod (i.e. the prior conc before perturbing xb)
  ! for monte carlo run for posterior error estimate comment the line below
  yobs=ymod

  ! Generate vector of random numbers of length equal to xb
  
  call random_seed
  call random_seed(size=ssize)
  allocate(seedvec(ssize))
  do i=1,ssize
    seedvec(i)=seed
  end do
  call random_seed(put=seedvec(1:ssize))
  do i=1,nn
    call random_number(ri)
    rbvec(i)=ri
  end do
  do i=1,nm
    call random_number(ri)
    rrvec(i)=ri
  end do
  deallocate(seedvec)

  ! normalize random numbers and subtract 1 to have mean of zero
  rrvec=rrvec*nm/sum(rrvec)-1.
  rbvec=rbvec*nn/sum(rbvec)-1.

  write(seedstr,'(I3)') seed
  open(100,file=trim(dirinvert)//'rand'//trim(seedstr)//'.txt',action='write',status='replace',iostat=ierr)
  do i=1,nn
    write(100,fmt='(F6.4)') anint(rbvec(i)*10000.)/10000.
  end do
  close(100)

  ! Calculate perturbation B^(1/2)*r
  ! note: no perturbation is added to the scalars of the background

  ! TESTING 
  sqrtb(:,:)=0.
  !

  ! uses B^(1/2) = Q*L^(1/2)*Q^T
  do jt=1,nemsteps
    do i=1,nbox
      do j=1,neig
        tmp(1,j)=bevec(i,j)*sqrt(beval(j))
      end do
      do k=1,nbox
        tmp1(1:1,k:k)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
  
        ! TESTING 
        sqrtb(i,k)=tmp1(1,k)
        ! 

      end do
      do it=1,nemsteps 
        xb((jt-1)*nbox+i:(jt-1)*nbox+i)=xb((jt-1)*nbox+i:(jt-1)*nbox+i)+&
            &cortime(it,jt)*matmul(real(tmp1,kind=4),rbvec((it-1)*nbox+1:it*nbox))
      end do
    end do
  end do
  
  ! Calculate perturbation of R^(0.5)*r

  ! calculate R^(0.5) by first computing eigenvalue decomposition
  N=nm
  lda=nm
  ldz=nm
  U=dble(covr)
  vl=1.
  vu=1.
  il=1
  iu=1
  abstol=0.
  lwork=26*N
  liwork=10*N
  allocate(work(lwork))
  allocate(iwork(liwork))
  allocate(isuppz(2*N))
  allocate(Z(N,N))
  allocate(sqrtw(N,N))
  allocate(W(N))
  call dsyevr('V','A','L',N,U,lda,vl,vu,il,iu,abstol,M,W,Z,ldz,isuppz,work,lwork,iwork,liwork,info)
  if(info.ne.0) then
    write(logid,*) 'Error in eigenvalue decomposition'
    if(info.gt.0) write(logid,*) 'Internal error'
    if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
    stop
  endif
  deallocate(work)
  deallocate(iwork)
  deallocate(isuppz)

  ind=minloc(W,dim=1,mask=W.gt.0.)
  sqrtw(:,:)=0.
  do i=ind,nm
    sqrtw(i,i)=sqrt(W(i))
  end do
  sqrtcovr(:,ind:nm)=matmul(Z(:,ind:nm),sqrtw(ind:nm,ind:nm))
  sqrtcovr=matmul(sqrtcovr(:,ind:nm),transpose(Z(:,ind:nm)))
  deallocate(sqrtw)
  deallocate(Z)
  deallocate(W)

  ! for synthetic data test add random noise to observations
  ! for monte carlo run for posterior error estimate comment the line below
  yobs(1:nm)=yobs(1:nm)+matmul(real(sqrtcovr,kind=4),rrvec)


!  do i=1,nm
!    yobs(i)=yobs(i)+sqrt(covr(i,i))*rrvec(i)
!  end do

  !--------------------------------------------------------------
  ! TESTING: for testing eigenvalue calculations
  !--------------------------------------------------------------
  ! B^(1/2) (sqrtb)
  fileout='sqrt_eigen.nc'
  varname='sqrtb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='sqrt of B reconstructed from eigendecomposition'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(sqrtb,4),&
                   &nbox,nbox,dimbox,dimbox,dname1,dname2,comment)
  ! xb (perturbed prior emissions)
  open(300,file=trim(dirinvert)//'prior_perturb.txt',status='replace',action='write',iostat=ierr)
  do i=1,nbox*nemsteps
    write(300,*) xb(i)/numscale
  end do
  do i=1,nbkgnd
    write(300,*) xb(nbox*nemsteps+i)
  end do
  close(300) 
  ! yobs (perturbed observations)
  open(300,file=trim(dirinvert)//'yobs_perturb.txt',status='replace',action='write',iostat=ierr)
  do i=1,nm
    write(300,*) yobs(i)
  end do
  close(300)
  !--------------------------------------------------------------

end subroutine calcperturb


