!---------------------------------------------------------------------------------------
! FLEXINVERT : calcsrmatrix
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculates the source-receptor matrix describing the relationship between 
!! surface fluxes and the observations at a given receptor for the nested domain.
!! Reads the surface influence grid_time files for each station and timestep and
!! allocates these to the corresponding row and columns of the source-receptor matrix.
!! Also calculates contribution from each grid-cell to the background mixing ratio 
!! for each observation and converts this to the coarse grid resolution.
!---------------------------------------------------------------------------------------

subroutine calcsrmatrix(knest,kout)

  use mod_var
  use mod_dates

  implicit none

  character(len=100) :: dirflexrec,filename,datestr,timestr
  integer :: ind,indo,indm,inde,cnt,cntp,numpoint
  integer :: i,nr,nt,k,ix,jy
  logical :: nextmonth,nextemstep
  integer :: yyyymmddhh,yyyymm,preyyyymm,yyyy,mm,eomday
  real(kind=8) :: jd,jdi,jdf,jdo
  real, dimension(maxrows,nemsteps*nbox) :: knest
  real, dimension(maxrows,ncrs) :: kout
  real, dimension(nzmax) :: outheight
  integer, dimension(maxpoint) :: releases
  real, dimension(:,:,:,:), allocatable :: grid,mgrid,mgridp
  real, dimension(:), allocatable :: surfinfvec,surfinfvecp,surfinfvecfin,bkgndinf,bkgndinfcrs
  real, dimension(:), allocatable :: mgridvec,mgridvecp
  real :: jtime
  integer :: count_i,count_f,count_rate

  allocate(indexobs(maxrows,2))
  allocate(grid(numxgrid,numygrid,numzlev,maxpoint))
  allocate(mgrid(numxgrid,numygrid,numzlev,maxpoint))
  allocate(mgridp(numxgrid,numygrid,numzlev,maxpoint))  

  allocate(surfinfvec(nbox))
  allocate(surfinfvecp(nbox))
  allocate(surfinfvecfin(numxnest*numynest))
  allocate(bkgndinf(numxgrid*numygrid))
  allocate(bkgndinfcrs(ncrs))
  allocate(mgridvec(numxgrid*numygrid))
  allocate(mgridvecp(numxgrid*numygrid))

  mgrid(:,:,:,:)=0.
  mgridp(:,:,:,:)=0.

  write(logid,*) 'Building knest and kout matrices...'

  indm=0
  ! loop over receptors
  do nr=1,numrec
    indo=0
    jd=juldatei
    jdi=juldatei
    inde=1
    nextmonth=.true.
!    print*, 'juldatei:',juldatei
    write(logid,*) 'Reading grids for receptor: '//trim(recname(nr))
    ! loop over trajectories month by month and allocate surface influence
    ! to relevant emission timestep
    do while (jd.le.juldatef)
      jdo=jd
      yyyymmddhh=caldate(jd)
      yyyymm=yyyymmddhh/10000
      eomday=calceomday(yyyymm)
      preyyyymm=yyyymm
!      print*, 'yyyymm:',yyyymm      
      write(logid,fmt='(1X,A24,1X,I6)') 'Reading grids for month:',preyyyymm
      write(datestr,'(i6)') yyyymm
      dirflexrec=trim(dirflex)//trim(recname(nr))//'/'//trim(datestr)//'/'
      ! if using flexpart option 'SURF_ONLY'
      filename=trim(dirflexrec)//'header_grid_time'
      ! if not using option 'SURF_ONLY'
!      filename=trim(dirflexrec)//'header'
      call readheader(filename,numpoint,outheight,releases)
      ! if in this directory for first time then adjust
      ! start date for trajectories going back in time
      if(nextmonth) jd=max(juldatei,(jd-trajlen))
        do while (jd.lt.(jdi+tstep))
          call readgrid(dirflexrec,jd,numpoint,grid)
          yyyymmddhh=caldate(jd)
          write(timestr,'(i10)') yyyymmddhh
!          print*, 'reading: ',trim(dirflexrec)//'grid_time_'//trim(timestr)//'0000_001'
          if(jd.lt.jdi.and.jd.gt.(jdi-tstep)) then
            ! grid belongs to previous timestep
            mgridp=mgridp+grid
            cntp=cntp+1
          elseif(jd.ge.jdi) then
            ! grids belong to current timestep
            mgrid=mgrid+grid
            cnt=cnt+1
          endif
          jd=jd+float(abs(loutstep))/3600./24.
          yyyymmddhh=caldate(jd)
          yyyymm=yyyymmddhh/10000
          ! if not finished emission timestep but need to move to next month
          if(yyyymm.gt.preyyyymm) go to 20
        end do ! jd
20    continue
!      print*, 'numpoint:',numpoint
!      print*, 'jd:',jd
      ! assign knest and kout matrices by looping over trajectories in emission timestep
      ! keep original trajectory time resolution for now
      ! knest and kout organised by receptor, month, and trajectory
      do k=1,numpoint
        ind=indm+k
        if(nextmonth) then
          indo=indo+1
          ! record index value for comparison to observations
          indexobs(ind,1)=nr
          if(trim(rectyp(nr)).eq.'FM'.or.trim(rectyp(nr)).eq.'CM') then
            ! index by model time step in hours
            indexobs(ind,2)=int((jdo-juldatei)*24+k*relperiod)
          else
            ! index by number of observations 
            ! note: assumes one back trajectory for each observation
            indexobs(ind,2)=indo
          endif
        endif
!        print*, 'ind:',ind
!        print*, 'indexobs:',indexobs(ind,2)
        ! vectorize surface influence and calculate contribution to
        ! background from outside nested domain
        yyyy=yyyymmddhh/1000000
        mm=(yyyymmddhh-yyyy*1000000)/10000
        ! note: releases gives hour of trajectory since 1st release in backwards time
        jtime=float(yyyy)+float(mm-1)/12.+float(eomday*24+releases(k)/3600)/24./365.
!        print*, 'jtime:',jtime
!        if(k.eq.1) print*, 'timemi:',timemi 
        nt=minloc(abs(timemi(:)-jtime),dim=1,mask=abs(timemi(:)-jtime).eq.minval(abs(timemi(:)-jtime)))
!        print*, 'nt:',nt
        mgridvec(:)=0.
        mgridvecp(:)=0.
        bkgndinf(:)=0.
        call system_clock(count_i,count_rate)
        do jy=1,numygrid
          do ix=1,numxgrid
            mgridvec((jy-1)*numxgrid+ix)=mgrid(ix,jy,1,k)
            mgridvecp((jy-1)*numxgrid+ix)=mgridp(ix,jy,1,k)
            ! calculate contribution to change in mixing ratio from outside domain
            ! divide emisflex by numscale
            bkgndinf((jy-1)*numxgrid+ix)=mgrid(ix,jy,1,k)*emisflex(ix,jy,nt)*area_grid(ix,jy)/numscale
          end do
        end do
        call system_clock(count_f,count_rate)
!        print*, 'calcsrmatrix sec in loop1:',real(count_f-count_i+1)/real(count_rate)
        ! surface influence for nested domain at fine resolution and for
        ! contribution to background set values inside nest to zero 
        surfinfvecfin(:)=0.
        call system_clock(count_i,count_rate)
        do jy=1,numynest
          do ix=1,numxnest
            surfinfvecfin((jy-1)*numxnest+ix)=mgrid(ixl+ix-1,iyl+jy-1,1,k)
            bkgndinf((iyl+jy-2)*numxgrid+ixl+ix-1)=0.
          end do
        end do
        call system_clock(count_f,count_rate)
!        print*, 'calcsrmatrix sec in loop2:',real(count_f-count_i+1)/real(count_rate)
        ! map surface influence to inversion grid in nested domain
        ! no area weighting of surfinf because emissions already account for area
        surfinfvec=matmul(real(gridoper),mgridvec)
        surfinfvecp=matmul(real(gridoper),mgridvecp)
        if(calcbg) then 
          ! map contribution to background to coarse grid
          bkgndinfcrs=matmul(real(gridcrs),bkgndinf)
          bkgndinfcrs=bkgndinfcrs/area_crs
        endif
        if(.not.lowmemory) then
          ! write to knest_fin 
          ! divide by outheight(1) to give surface influence in s m2/kg
          knest_fin(ind,:)=knest_fin(ind,:)+surfinfvecfin/outheight(1)
        endif
        ! write to knest
        ! divide by outheight(1) to give surface influence in s m2/kg
        ! write contribution to previous emission timestep if any
        if((inde-1).gt.0.and.cntp.gt.0) knest(ind,((inde-2)*nbox+1):(inde-1)*nbox)=&
                                              &knest(ind,((inde-2)*nbox+1):(inde-1)*nbox)+surfinfvecp/outheight(1)
        ! write contribution to current emission timestep if any
        if(inde.gt.0.and.cnt.gt.0) knest(ind,((inde-1)*nbox+1):inde*nbox)=&
                                              &knest(ind,((inde-1)*nbox+1):inde*nbox)+surfinfvec/outheight(1)
        ! write to kout
        ! divide by outheight(1) to give mixing ratios
        if(calcbg) kout(ind,:)=bkgndinfcrs/outheight(1)
      end do ! numpoint 
      if(yyyymm.gt.preyyyymm) then
        ! move to next months files
        nextmonth=.true.
        indm=indm+numpoint
        if(jd.ge.(jdi+tstep)) then
           ! also new emission timestep
           inde=inde+1
           jdi=jdi+tstep
        endif
      else
        ! stay in same month but move to next emission timestep
        nextmonth=.false.
        if(jd.ge.(jdi+tstep)) then
           inde=inde+1
           jdi=jdi+tstep
        endif
      endif
!      print*, 'nextmonth:',nextmonth
!      print*, 'inde:',inde
!      print*, 'indm:',indm
      if(inde.gt.nemsteps) go to 30
      mgridp(:,:,:,:)=0.
      mgrid(:,:,:,:)=0.
      cntp=0
      cnt=0
    end do ! juldatef
30  continue
    ! finished current receptor move to next 
    if(.not.nextmonth) indm=indm+numpoint
!    print*, 'indm:',indm
    mgridp(:,:,:,:)=0.
    mgrid(:,:,:,:)=0.
    cntp=0
    cnt=0
  end do ! numrec

  if(.not.lowmemory) then
    ! mean over nemsteps
    knest_fin=knest_fin/nemsteps
  endif

  ! convert from ppt mass mixing ratio to volume mixing ratio
  knest=knest*coef*28.97/molarmass
  kout=kout*coef*28.97/molarmass
  if(.not.lowmemory) then
    knest_fin=knest_fin*coef*28.97/molarmass
  endif

  deallocate(grid)
  deallocate(mgrid)
  deallocate(mgridp)
  deallocate(surfinfvec)
  deallocate(surfinfvecp)
  deallocate(surfinfvecfin)
  deallocate(mgridvec)
  deallocate(mgridvecp)
  deallocate(bkgndinf)
  deallocate(bkgndinfcrs)

  open(600,file=trim(dirinvert)//'indexobs.txt',status='replace',action='write')
  do i=1,ind
    write(600,fmt='(I2,1X,I5)') indexobs(i,1),indexobs(i,2)
  end do
  close(600)

  open(800,file=trim(dirinvert)//'knest.txt',status='replace',action='write')
  do i=1,ind
    write(800,*) (knest(i,k),k=1,nbox*nemsteps)
  enddo
  close(800)

end subroutine calcsrmatrix

