!---------------------------------------------------------------------------------------
! FLEXINVERT : erroraggreg
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculates the error in the observation space due to the spatial aggregation of the
!! inversion grid:
!!
!! Ragg = HPBP^TH^T
!!
!! where:
!!   P  = projection operator for loss of information in coarse grid (nxn)
!!   B  = prior error covariance matrix (nxn)
!!   H  = kernel matrix (mxn)
!!
!! and P is calculated as:
!!
!! P = I - sum(pi*pi^T)
!!
!! where pi is an orthogonal vector of unit length and is the mapping of the fine grid
!! to each grid cell on the coarse grid
!!
!! Ref: Kaminski et al., J.Geophys.Res., 2001
!! note: H and B are on the fine resolution grid
!---------------------------------------------------------------------------------------

subroutine erroraggreg(nn,np,nm,nw,covagg)

  use mod_var

  implicit none

  integer :: nn,np,nm,nw,i,j,k,r
  real(kind=8), dimension(nm,nm) :: covagg
  real(kind=8), dimension(nm,nn) :: kernp,kernpb
  real(kind=8), dimension(nn,nm) :: tkernp
  real, dimension(nn,1) :: pvec
  real(kind=8), dimension(nm,1) :: kpvec
  real(kind=8), dimension(1,neigfin) :: tmp
  real(kind=8), dimension(1,nn) :: tmp1
  
  ! calculate HP via H-H*sum(pi*pi^T)
  kernp(:,:)=0.
  do i=1,nw
    ! vector pi is unit vector
    pvec(:,1)=real(gridnest(i,:))/sqrt(real(sum(gridnest(i,:))))
    kpvec(:,1)=matmul(dble(knest_fintrim(1:nm,:)),dble(pvec(:,1)))
    kernp=kernp+matmul(kpvec,transpose(dble(pvec)))
  enddo
  kernp=dble(knest_fintrim(1:nm,:))-kernp  

  ! calculate HPB
  ! use eigendecomposition of B
  do r=1,nn
    do i=1,nn
      do j=1,neigfin
        tmp(1:1,j:j)=bfinevec(i,j)*bfineval(j)
      end do
      tmp1(1:1,i:i)=matmul(tmp,reshape(bfinevec(r,:),(/neigfin,1/)))
    end do
    do k=1,nm
      kernpb(k:k,r)=matmul(kernp(k,:),reshape(tmp1,(/nn,1/)))
    end do
  end do

  ! calculate HPB(HP)^T
  tkernp=transpose(kernp)
  covagg=matmul(kernpb,tkernp)

end subroutine erroraggreg

