!---------------------------------------------------------------------------------------
! FLEXINVERT : errorcovariance
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculate the prior error covariance matrix using by first calculating the spatial
!! covariance at fine resolution, second averaging this to the coarse resolution and
!! lastly forming the temporal-spatial error covariance at coarse resolution.
!---------------------------------------------------------------------------------------

subroutine errorcovariance()

  use mod_var

  implicit none

  character(len=100) :: varname,varunit,comment
  character(len=20) :: dname1,dname2
  logical :: lexist 
  integer :: i,j,n,ik,jk
  integer :: ind,lda,lwork,liwork,info,M
  integer, dimension(:), allocatable :: isuppz
  real(kind=8), dimension(:), allocatable :: work
  integer, dimension(:), allocatable :: iwork
  real(kind=8), dimension(:), allocatable :: W
  real(kind=8), dimension(:,:), allocatable :: Z
  real :: covbsum,toterr
  real, dimension(:), allocatable :: xerr,xerrfin,xerrfinaw
  real, dimension(:,:), allocatable :: corspace
  real(kind=8), dimension(:,:), allocatable :: covbspace,covbtstep,lhs
  real, dimension(:,:), allocatable :: tmat,dtmat
  real, dimension(:), allocatable :: tsteps
  real, dimension(numxgrid*numygrid) :: area_vec
  real(kind=8), dimension(numxnest*numynest,numxnest*numynest) :: covbfin,covbfinaw

  !---------------------------------------------------------------
  ! calculate error covariance for a single timestep
  !---------------------------------------------------------------

  allocate(corspace(numxnest*numynest,numxnest*numynest))

  ! check for existing spatial correlation
  write(nxst,'(i3.3)') numxnest
  write(nyst,'(i3.3)') numynest
  write(sigst,'(i4.4)') int(sigmaland)
  filecorrel='correl_'//trim(nxst)//'x'//trim(nyst)//'_'//trim(sigst)//'.nc'
  inquire(file=trim(dirauxil)//trim(filecorrel),exist=lexist)

  if(.not.lexist) then
    ! calculate spatial correlation
    call correl(corspace)
    varname='correl'
    varunit='none'
    dname1='dimx'
    dname2='dimy'
    comment='spatial correlation'
    call write2Dncdf(dirauxil,filecorrel,&
                     &varname,varunit,corspace,numxnest*numynest,&
                     &numxnest*numynest,dimxynest,dimxynest,dname1,dname2,comment)
  else
    ! read existing spatial correlation
    filecorrel=trim(dirauxil)//'correl_'//trim(nxst)//'x'//trim(nyst)//'_'//trim(sigst)//'.nc'
    varname='correl'
    dname1='dimx'
    dname2='dimy'
    call read2Dncdf(filecorrel,varname,dname1,dname2,&
                    &numxnest*numynest,numxnest*numynest,corspace)
  endif

  allocate(covbspace(numxnest*numynest,numxnest*numynest))
  allocate(xerr(numxnest*numynest))
  allocate(xerrfin(numxnest*numynest))
  allocate(xerrfinaw(numxnest*numynest))

  ! calculate prior error vector 
  xerrfin(:)=0.
  xerrfinaw(:)=0.
  xerr(:)=0.
  do j=1,numynest
    do i=1,numxnest
      xerr((j-1)*numxnest+i)=errprior(ixl+i-1,iyl+j-1)*area_grid(ixl+i-1,iyl+j-1)/area_box(nbox_xy(i,j))
      xerrfinaw((j-1)*numxnest+i)=errprior(ixl+i-1,iyl+j-1)*area_grid(ixl+i-1,iyl+j-1)/area_box(nbox_xy(i,j))
      xerrfin((j-1)*numxnest+i)=errprior(ixl+i-1,iyl+j-1)
    end do
  end do

  ! calculate error covariance at fine resolution
  area_vec(:)=0.
  do j=1,numynest
    do i=1,numxnest
      area_vec((j-1)*numxnest+i)=area_grid(ixl+i-1,iyl+j-1)
    end do
  end do
  covbspace(:,:)=0.
  covbsum=0.
  do j=1,numxnest*numynest
    do i=j,numxnest*numynest
      covbspace(i,j)=dble(corspace(i,j))*dble(xerr(i))*dble(xerr(j))
      covbspace(j,i)=covbspace(i,j)
      if(i.eq.j) covbsum=covbsum+covbspace(i,j)*area_vec(j)**2
      if(i.ne.j) covbsum=covbsum+2*covbspace(i,j)*area_vec(j)*area_vec(i)
    end do
  end do

  ! scale error to be consistent with total error for nested domain (in Tg/y)
  toterr=sqrt(covbsum)*3600.*24.*365./numscale/1.e9
  erscalar=globerr/toterr
  covbtstep=covbtstep*erscalar**2

  write(logid,fmt='(1X,A22,F9.3)') 'Original total error: ',toterr
  write(logid,fmt='(1X,A23,F9.3)') 'Total error scaled by: ',erscalar 

  ! map prior error covariances to inversion grid

  allocate(lhs(nbox,numxnest*numynest))
  allocate(covbtstep(nbox,nbox))

  lhs=matmul(dble(gridnest),covbspace)
  deallocate(covbspace)
  covbtstep=matmul(lhs,transpose(dble(gridnest)))
  deallocate(lhs)

  ! calculate eigendecomposition of covbtstep
  ! note: eigenvalues are sorted in ascending order
  lwork=26*nbox
  liwork=10*nbox
  allocate(work(lwork))
  allocate(iwork(liwork))
  allocate(isuppz(2*nbox))
  allocate(Z(nbox,nbox))
  allocate(W(nbox))
  call dsyevr('V','A','L',nbox,covbtstep,nbox,1.,1.,1,1,0.,M,W,Z,nbox,isuppz,work,lwork,iwork,liwork,info)
  if(info.ne.0) then
    write(logid,*) 'Error in eigenvalue decomposition'
    if(info.gt.0) write(logid,*) 'Internal error'
    if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
    stop
  endif
  ind=minloc(W,dim=1,mask=W.ge.trunc)
  write(logid,fmt='(1X,A26,F5.2)') 'Truncating eigenvalues at ',trunc
  neig=nbox-ind+1
  write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neig
  deallocate(covbtstep)
  allocate(beval(neig))
  allocate(bevec(nbox,neig))
  beval=W(ind:nbox)
  bevec=Z(:,ind:nbox)
  deallocate(work)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(W)
  deallocate(Z)

  !---------------------------------------------------------------
  ! calculate temporal correlation over all time steps
  !---------------------------------------------------------------

  allocate(cortime(nemsteps,nemsteps))
  allocate(icortime(nemsteps,nemsteps))
  allocate(tsteps(nemsteps))
  allocate(tmat(nemsteps,nemsteps))
  allocate(dtmat(nemsteps,nemsteps))

  ! calculate temporal correlation as expontential decay with time (days)

  do i=1,nemsteps
    tsteps(i)=i*tstep
  end do
  tmat=matmul(reshape(tsteps,(/nemsteps,1/)),reshape((tsteps/tsteps),(/1,nemsteps/)))
  dtmat=abs(tmat-transpose(tmat))

  do i=1,nemsteps
    do j=i,nemsteps
      cortime(i,j)=exp(-dtmat(i,j)/sigmatime)
      cortime(j,i)=exp(-dtmat(j,i)/sigmatime)
    end do
  end do

  ! calculate inverse of cortime
  allocate(Z(nemsteps,nemsteps))
  Z=dble(cortime) ! symmetric matrix
  lda=nemsteps    ! leading dimension of cortime
  call dpotrf('L',nemsteps,Z,lda,info)
  call dpotri('L',nemsteps,Z,lda,info)
  if(info.gt.0) write(logid,*) 'Error in inverting cortime'
  ! returns lower ('L') triangle of inverse of cortime
  icortime=Z
  do i=1,nemsteps
    do j=1,i
      icortime(j,i)=Z(i,j)
    end do
  end do
  deallocate(Z)

  !---------------------------------------------------------------
  ! calculate error covariance matrix on fine grid
  !---------------------------------------------------------------

  ! used in calculation of aggregation error 

  covbfin(:,:)=0.
  covbfinaw(:,:)=0.
  do j=1,numxnest*numynest
    do i=j,numxnest*numynest
      covbfin(i,j)=dble(corspace(i,j))*dble(xerrfin(i))*dble(xerrfin(j))
      covbfinaw(i,j)=dble(corspace(i,j))*dble(xerrfinaw(i))*dble(xerrfinaw(j))
      covbfin(j,i)=covbfin(i,j)
      covbfinaw(j,i)=covbfinaw(i,j)
    end do
  end do
  deallocate(corspace)

  ! scale error to be consistent with global error
  covbfin=covbfin*(globerr/toterr)**2
  covbfinaw=covbfinaw*(globerr/toterr)**2

  ! calculate eigendecomposition of covbfin
  ! note: eigenvalues are sorted in ascending order
  lwork=26*numxnest*numynest
  liwork=10*numxnest*numynest
  allocate(work(lwork))
  allocate(iwork(liwork))
  allocate(isuppz(2*numxnest*numynest))
  allocate(Z(numxnest*numynest,numxnest*numynest))
  allocate(W(numxnest*numynest))
  call dsyevr('V','A','L',numxnest*numynest,covbfin,numxnest*numynest,1.,1.,1,1,0.,&
              &M,W,Z,numxnest*numynest,isuppz,work,lwork,iwork,liwork,info)
  if(info.ne.0) then
    write(logid,*) 'Error in eigenvalue decomposition'
    if(info.gt.0) write(logid,*) 'Internal error'
    if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
    stop
  endif
  ind=minloc(W,dim=1,mask=W.ge.trunc)
  write(logid,fmt='(1X,A26,F5.2)') 'Truncating eigenvalues at ',trunc
  neigfin=numxnest*numynest-ind+1
  write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neigfin
  allocate(bfineval(neigfin))
  allocate(bfinevec(numxnest*numynest,neigfin))
  bfineval=W(ind:numxnest*numynest)
  bfinevec=Z(:,ind:numxnest*numynest)
  deallocate(work)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(W)
  deallocate(Z)

  ! calculate eigendecomposition of covbfinaw
  ! note: eigenvalues are sorted in ascending order
  lwork=26*numxnest*numynest
  liwork=10*numxnest*numynest
  allocate(work(lwork))
  allocate(iwork(liwork))
  allocate(isuppz(2*numxnest*numynest))
  allocate(Z(numxnest*numynest,numxnest*numynest))
  allocate(W(numxnest*numynest))
  call dsyevr('V','A','L',numxnest*numynest,covbfinaw,numxnest*numynest,1.,1.,1,1,0.,&
              &M,W,Z,numxnest*numynest,isuppz,work,lwork,iwork,liwork,info)
  if(info.ne.0) then
    write(logid,*) 'Error in eigenvalue decomposition'
    if(info.gt.0) write(logid,*) 'Internal error'
    if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
    stop
  endif
  ind=minloc(W,dim=1,mask=W.ge.trunc)
  write(logid,fmt='(1X,A26,F5.2)') 'Truncating eigenvalues at ',trunc
  neigfinaw=numxnest*numynest-ind+1
  write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neigfinaw
  allocate(bfinaweval(neigfinaw))
  allocate(bfinawevec(numxnest*numynest,neigfinaw))
  bfinaweval=W(ind:numxnest*numynest)
  bfinawevec=Z(:,ind:numxnest*numynest)
  deallocate(work)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(W)
  deallocate(Z)

  ! prepare to exit
  deallocate(tsteps)
  deallocate(tmat)
  deallocate(dtmat)
  deallocate(xerr)
  deallocate(xerrfin)
  deallocate(xerrfinaw)

end subroutine errorcovariance





