!---------------------------------------------------------------------------------------
! FLEXINVERT : errorcovariance_coarse
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Calculate the prior error covariance matrix using by first calculating the spatial
!! covariance at coarse resolution, and second forming the temporal-spatial error 
!! covariance.
!---------------------------------------------------------------------------------------

subroutine errorcovariance_coarse()

  use mod_var

  implicit none

  character(len=100) :: fileout,varname,varunit,comment
  character(len=20) :: dname1,dname2
  logical :: lexist 
  integer :: i,j,n,ik,jk,ierr
  integer :: ind,lda,lwork,liwork,info,M
  integer, dimension(:), allocatable :: isuppz
  real(kind=8), dimension(:), allocatable :: work
  integer, dimension(:), allocatable :: iwork
  real(kind=8), dimension(:), allocatable :: W
  real(kind=8), dimension(:,:), allocatable :: Z
  real :: toterr,covbsum
  real, dimension(:), allocatable :: xerr,xerrfin,xerrfinaw
  real, dimension(:,:), allocatable :: corspace
  real(kind=8), dimension(:,:), allocatable :: covbtstep,icovbtstep,tmp
  real, dimension(:,:), allocatable :: tmat,dtmat
  real, dimension(:), allocatable :: tsteps
  real(kind=8), dimension(:,:), allocatable :: covbfin,covbfinaw

  !---------------------------------------------------------------
  ! calculate error covariance for a single timestep
  !---------------------------------------------------------------

  if(.not.lowmemory.and.scaleflx.eq.2) then
    allocate(covbfin(numxnest*numynest,numxnest*numynest))
    allocate(covbfinaw(numxnest*numynest,numxnest*numynest))
  endif

  allocate(corspace(nbox,nbox))

  ! check for existing spatial correlation
  write(nxst,'(I4.4)') nbox
  write(sigst,'(I4.4)') int(sigmaland)
  filecorrel='correl_'//trim(nxst)//'_'//trim(sigst)//'.nc'
  inquire(file=trim(dirauxil)//trim(filecorrel),exist=lexist)

  if(.not.lexist) then
    ! calculate spatial correlation
    call correl_coarse(corspace)
    varname='correl'
    varunit='none'
    dname1='dimx'
    dname2='dimy'
    comment='spatial correlation'
    call write2Dncdf(dirauxil,filecorrel,&
                     &varname,varunit,corspace,nbox,&
                     &nbox,dimbox,dimbox,dname1,dname2,comment)
  else
    ! read existing spatial correlation
    filecorrel=trim(dirauxil)//'correl_'//trim(nxst)//'_'//trim(sigst)//'.nc'
    varname='correl'
    dname1='dimx'
    dname2='dimy'
    call read2Dncdf(filecorrel,varname,dname1,dname2,&
                    &nbox,nbox,corspace)
  endif

  allocate(covbtstep(nbox,nbox))
  allocate(xerr(nbox))

  if(.not.lowmemory) then
    allocate(xerrfin(numxnest*numynest))
    allocate(xerrfinaw(numxnest*numynest))
  ! calculate prior error vector for nest at fine resolution 
    xerrfin(:)=0.
    xerrfinaw(:)=0.
    do j=1,numynest
      do i=1,numxnest
        xerrfinaw((j-1)*numxnest+i)=errprior(ixl+i-1,iyl+j-1)*area_grid(ixl+i-1,iyl+j-1)/area_box(nbox_xy(i,j))
        xerrfin((j-1)*numxnest+i)=errprior(ixl+i-1,iyl+j-1)
      end do
    end do
  endif

  ! calculate prior error vector for nest at coarse resolution 
  xerr(:)=0.
  do j=1,numynest
    do i=1,numxnest
      xerr(nbox_xy(i,j))=xerr(nbox_xy(i,j))+errprior(ixl+i-1,iyl+j-1)*area_grid(ixl+i-1,iyl+j-1)/area_box(nbox_xy(i,j))
    end do
  end do

  ! calculate error covariance at coarse resolution
  covbtstep(:,:)=0.
  covbsum=0.
  do j=1,nbox
    do i=1,nbox
      covbtstep(i,j)=dble(corspace(i,j))*dble(xerr(i))*dble(xerr(j))
      covbsum=covbsum+covbtstep(i,j)*area_box(j)*area_box(i)
    end do 
  end do
  deallocate(corspace)

  ! scale error to be consistent with total error of nested domain (in Tg/y)
  toterr=sqrt(covbsum)*3600.*24.*365./numscale/1.e9
  erscalar=globerr/toterr
  covbtstep=covbtstep*erscalar**2

  ! TESTING 
  fileout='covb_orig.nc'
  varname='covb'
  varunit='none'
  dname1='dimx'
  dname2='dimy'
  comment='covb original'
  call write2Dncdf(dirinvert,fileout,varname,varunit,real(covbtstep,4),&
                   &nbox,nbox,dimbox,dimbox,dname1,dname2,comment)
  !

  write(logid,fmt='(1X,A22,F10.6)') 'Original total error: ',toterr
  write(logid,fmt='(1X,A23,F10.6)') 'Total error scaled by: ',erscalar

  ! calculate eigendecomposition of covbtstep
  ! note: eigenvalues are sorted in ascending order
  lwork=26*nbox       
  liwork=10*nbox    
  allocate(work(lwork))
  allocate(iwork(liwork))
  allocate(isuppz(2*nbox))
  allocate(Z(nbox,nbox))
  allocate(W(nbox))
  call dsyevr('V','A','L',nbox,covbtstep,nbox,1.,1.,1,1,0.,M,W,Z,nbox,isuppz,work,lwork,iwork,liwork,info)
  if(info.ne.0) then
    write(logid,*) 'Error in eigenvalue decomposition'
    if(info.gt.0) write(logid,*) 'Internal error'
    if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
    stop
  endif
  ind=minloc(W,dim=1,mask=W.ge.(trunc*maxval(W)))
  write(logid,fmt='(1X,A26,F10.2)') 'Truncating eigenvalues at ',trunc*maxval(W)
  neig=nbox-ind+1
  write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neig
  deallocate(covbtstep)
  allocate(beval(neig))
  allocate(bevec(nbox,neig))
  beval=W(ind:nbox)
  bevec=Z(:,ind:nbox)
  deallocate(work)
  deallocate(iwork)
  deallocate(isuppz)
  deallocate(W)
  deallocate(Z)

  !---------------------------------------------------------------
  ! calculate temporal correlation over all time steps
  !---------------------------------------------------------------

  allocate(cortime(nemsteps,nemsteps))
  allocate(icortime(nemsteps,nemsteps))
  allocate(tsteps(nemsteps))
  allocate(tmat(nemsteps,nemsteps))
  allocate(dtmat(nemsteps,nemsteps))

  ! calculate temporal correlation as expontential decay with time (days)

  do i=1,nemsteps
    tsteps(i)=i*tstep
  end do
  tmat=matmul(reshape(tsteps,(/nemsteps,1/)),reshape((tsteps/tsteps),(/1,nemsteps/)))
  dtmat=abs(tmat-transpose(tmat))

  do i=1,nemsteps
    do j=i,nemsteps
      cortime(i,j)=exp(-dtmat(i,j)/sigmatime)
      cortime(j,i)=exp(-dtmat(j,i)/sigmatime)
    end do
  end do

  ! TESTING
  open(800,file=trim(dirinvert)//'cortime.txt',status='replace',action='write')
  do i=1,nemsteps
    write(800,*) (cortime(i,j),j=1,nemsteps)
  enddo
  close(800)
  !

  ! calculate inverse of cortime
  allocate(Z(nemsteps,nemsteps))
  Z=dble(cortime) ! symmetric matrix
  lda=nemsteps    ! leading dimension of cortime
  call dpotrf('L',nemsteps,Z,lda,info)
  call dpotri('L',nemsteps,Z,lda,info)
  if(info.gt.0) write(logid,*) 'Error in inverting cortime'
  ! returns lower ('L') triangle of inverse of cortime
  icortime=Z
  do i=1,nemsteps
    do j=1,i
      icortime(j,i)=Z(i,j)
    end do
  end do
  deallocate(Z)

  !---------------------------------------------------------------
  ! calculate error covariance matrix on fine grid
  !---------------------------------------------------------------

  if(.not.lowmemory) then

    ! used in calculation of aggregation error

    allocate(corspace(numxnest*numynest,numxnest*numynest))

    ! check for existing spatial correlation
    write(nxst,'(I3.3)') numxnest
    write(nyst,'(I3.3)') numynest
    filecorrel='correl_'//trim(nxst)//'x'//trim(nyst)//'_'//trim(sigst)//'.nc'
    inquire(file=trim(dirauxil)//trim(filecorrel),exist=lexist)

    if(.not.lexist) then
     ! calculate spatial correlation
      call correl(corspace)
      varname='correl'
      varunit='none'
      dname1='dimx'
      dname2='dimy'
      comment='spatial correlation.'
      call write2Dncdf(dirauxil,filecorrel,&
                       &varname,varunit,corspace,numxnest*numynest,&
                       &numxnest*numynest,dimxynest,dimxynest,dname1,dname2,comment)
    else
     ! read existing spatial correlation
      filecorrel=trim(dirauxil)//'correl_'//trim(nxst)//'x'//trim(nyst)//'_'//trim(sigst)//'.nc'
      varname='correl'
      dname1='dimx'
      dname2='dimy'
      call read2Dncdf(filecorrel,varname,dname1,dname2,&
                      &numxnest*numynest,numxnest*numynest,corspace)
    endif

    covbfin(:,:)=0.
    if(scaleflx.eq.2) covbfinaw(:,:)=0.

    do j=1,numxnest*numynest
      do i=j,numxnest*numynest
       ! error occurs in read/write of corspace, set diagonal to 1 to be sure
        if(i.eq.j) corspace(j,j)=1.
        if(scaleflx.eq.2) covbfinaw(i,j)=dble(corspace(i,j))*dble(xerrfinaw(i))*dble(xerrfinaw(j))
        covbfin(i,j)=dble(corspace(i,j))*dble(xerrfin(i))*dble(xerrfin(j))
        if(scaleflx.eq.2) covbfinaw(j,i)=covbfinaw(i,j)
        covbfin(j,i)=covbfin(i,j)
        if(i.eq.j.and.covbfin(j,j).lt.0.) then
          write(logid,fmt='(A30,1X,I6)') 'warning: covbfin non-positive at:',j
        endif
      end do
    end do
    deallocate(corspace)

    ! scale error to be consistent with global error
    covbfin=covbfin*(globerr/toterr)**2
    if(scaleflx.eq.2) covbfinaw=covbfinaw*(globerr/toterr)**2

   ! calculate eigendecomposition of covbfin
   ! note: eigenvalues are sorted in ascending order
    lwork=26*numxnest*numynest
    liwork=10*numxnest*numynest
    allocate(work(lwork))
    allocate(iwork(liwork))
    allocate(isuppz(2*numxnest*numynest))
    allocate(Z(numxnest*numynest,numxnest*numynest))
    allocate(W(numxnest*numynest))
    call dsyevr('V','A','L',numxnest*numynest,covbfin,numxnest*numynest,1.,1.,1,1,0.,&
                &M,W,Z,numxnest*numynest,isuppz,work,lwork,iwork,liwork,info)
    if(info.ne.0) then
      write(logid,*) 'Error in eigenvalue decomposition'
      if(info.gt.0) write(logid,*) 'Internal error'
      if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
      stop
    endif
    ind=minloc(W,dim=1,mask=W.ge.(trunc*maxval(W)))
    write(logid,fmt='(1X,A26,F10.2)') 'Truncating eigenvalues at ',trunc*maxval(W)
    neigfin=numxnest*numynest-ind+1
    write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neigfin
    allocate(bfineval(neigfin))
    allocate(bfinevec(numxnest*numynest,neigfin))
    bfineval=W(ind:numxnest*numynest)
    bfinevec=Z(:,ind:numxnest*numynest)
    deallocate(work)
    deallocate(iwork)
    deallocate(isuppz)
    deallocate(W)
    deallocate(Z)

    ! If using option to reoptimize to fine grid (scaleflx = 2)

    if(scaleflx.eq.2) then
     ! calculate eigendecomposition of covbfinaw
     ! note: eigenvalues are sorted in ascending order
      lwork=26*numxnest*numynest
      liwork=10*numxnest*numynest
      allocate(work(lwork))
      allocate(iwork(liwork))
      allocate(isuppz(2*numxnest*numynest))
      allocate(Z(numxnest*numynest,numxnest*numynest))
      allocate(W(numxnest*numynest))
      call dsyevr('V','A','L',numxnest*numynest,covbfinaw,numxnest*numynest,1.,1.,1,1,0.,&
                &M,W,Z,numxnest*numynest,isuppz,work,lwork,iwork,liwork,info)
      if(info.ne.0) then
        write(logid,*) 'Error in eigenvalue decomposition'
        if(info.gt.0) write(logid,*) 'Internal error'
        if(info.lt.0) write(logid,fmt='(1X,A25,I5,A21)') 'The argument in position ',info,' had an illegal value'
        stop
      endif
      ind=minloc(W,dim=1,mask=W.ge.(trunc*maxval(W)))
      write(logid,fmt='(1X,A26,F5.2)') 'Truncating eigenvalues at ',trunc*maxval(W)
      neigfinaw=numxnest*numynest-ind+1
      write(logid,fmt='(1X,A31,I5)') 'Number of retained eigenvalues ',neigfinaw
      allocate(bfinaweval(neigfinaw))
      allocate(bfinawevec(numxnest*numynest,neigfinaw))
      bfinaweval=W(ind:numxnest*numynest)
      bfinawevec=Z(:,ind:numxnest*numynest)
      deallocate(work)
      deallocate(iwork)
      deallocate(isuppz)
      deallocate(W)
      deallocate(Z)
    endif

  endif ! if not lowmemory

  !---------------------------------------------------------------
  ! prepare to exit
  !---------------------------------------------------------------

  deallocate(tsteps)
  deallocate(tmat)
  deallocate(dtmat)
  deallocate(xerr)
  if(.not.lowmemory) then
    deallocate(xerrfin)
    deallocate(xerrfinaw)
  endif

end subroutine errorcovariance_coarse





