!---------------------------------------------------------------------------------------
! FLEXINVERT
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!
!> FLEXINVERT
!! Guide to use:
!!   - Assumes a separate FLEXPART run for each receptor, where receptors can be 
!!     stations, flights or ship cruises. In the case of flights or ships, then 
!!     the receptor name corresponds to the flight/ship/campaign number and the
!!     grid_time files must correspond to the observations. 
!!   - Example of the default directory structure for FLEXPART output: 
!!     /FLEXPART_OUTPUT/RECNAME/YYYYMM/
!!   - Observations are read in from a standard format (see readobs.f90) and 
!!     a list of receptors with observation type is required (in files.def 'receptfile')
!!     in the receptfile observations must be indicated as one of the following:
!!       1) FM: Flask Measurement
!!       2) CM: Continuous Measurement
!!       3) AM/SM: Aircraft or Ship Measurement
!!   - Assumes grid_time units are s m3/kg
!!   - A background calculation method is built-in but relies on the grid_initial files
!!     from FLEXPART and 3D global fields of mean representative mixing ratios. 
!!     Alternatively the user may subtract the background from the observations prior to
!!     running FLEXINVERT (this is controlled by the option 'calcbg' in 'control.def')
!!   - Note that it is most efficient to write only surface layer to grid_time files
!!     (in FLEXPART9.2 this is option SURF_ONLY=1 in the COMMAND file)
!!   - Please read the user's guide available from http://flexinvert.nilu.no
!!
!! Modifications:
!!   - FLEXINVERT_NEST_01: 
!!       - replaced covb,covbfin,covbfinaw with eigenvalues and eigenvectors
!!   - FLEXINVERT_NEST_02: 
!!       - changed reading of observations/flexpart files to include moving receptors
!!       - removed background calculation based on observations
!!       - changed units of fluxes from kg/m3/s to kg/m2/s and kernel from s m3/kg to 
!!         s m2/kg (no longer assumes every flexpart run has same vertical resolution)
!!       - removed creation of 'array temporaries' 
!!   - FLEXINVERT_NEST_03:
!!       - changed reading of observations to avoid opening/closing of files for each read
!!       - added flexpart file checking in initparam.f90
!!       - added "restart" possibility to pick-up a run after knest, kout and obsread 
!!         are written to file
!!       - added "lowmemory" mode to avoid calculation of large matrices (note: if using
!!         this mode the aggregation error is not calculated and the regridding of the
!!         posterior fluxes to the fine grid is based only on the prior information)
!!     
!---------------------------------------------------------------------------------------

program main

  use mod_dates
  use mod_var

  implicit none

  ! path and file variables
  character(len=100) :: filename

  ! time variables
  character(len=100) :: datestr
  integer :: yyyymm

  ! flexpart variables 
  integer :: numpoint
  real, dimension(:), allocatable :: outheight
  integer, dimension(:), allocatable :: releases

  ! inversion variables
  integer :: n,i,j,k,it,jt,nobs,nmod,ierr
  real :: err,cost
  real, dimension(:,:), allocatable :: surfinf,memisflex,surfinfnest,memisflexnest
  real, dimension(:,:), allocatable :: bkgnd_obs,bkgnd
  real, dimension(:), allocatable :: yobs,ymod,ybgi,yunc,ypos,errb,erra,ebgb,ebga
  real, dimension(:), allocatable :: timeout,xb,x
  real, dimension(:,:), allocatable :: knest,knest_obs,knest_trim,kout,kout_obs,kout_trim
  real, dimension(:,:), allocatable :: covr,cova
  real(kind=8), dimension(:,:), allocatable :: gain,covagg
  real(kind=8), dimension(:,:), allocatable :: tmp,tmp1
  integer, dimension(:), allocatable :: ydat
  character(len=7), dimension(:), allocatable :: yrec
  real, dimension(:,:), allocatable :: allobs 
  real(kind=8), dimension(:,:), allocatable :: modout  

  !---------------------------------------------------------------
  ! read inversion definitions
  !---------------------------------------------------------------
 
  call readdefinitions()

  call system('cp control.def  '//trim(dirinvert))
  call system('cp files.def  '//trim(dirinvert))

  !---------------------------------------------------------------
  ! read list of receptors/stations
  !---------------------------------------------------------------

  call readreclist()

  !---------------------------------------------------------------
  ! read flexpart header
  !---------------------------------------------------------------

  ! read header to get domain dimensions etc.
  ! read for only one receptor as this must be the same for all!

  yyyymm=datei/100
  write(datestr,'(i6)') yyyymm
  ! if using flexpart option 'SURF_ONLY'
  filename=trim(dirflex)//trim(recname(1))//'/'//trim(datestr)//'/header_grid_time'
  ! if not using option 'SURF_ONLY'
!  filename=trim(dirflex)//trim(recname(1))//'/'//trim(datestr)//'/header'

  allocate(outheight(nzmax))
  allocate(releases(maxpoint))

  call readheader(filename,numpoint,outheight,releases)

  ! note: numzgrid may be different for grid_time and grid_init files
  ! to avoid confusion assign variable for z-dimension in grid_time
  
  numzlev=numzgrid 

  !---------------------------------------------------------------
  ! initialize various parameters 
  !---------------------------------------------------------------

  call initparam()  

  !---------------------------------------------------------------
  ! calculate mean surface influence
  !---------------------------------------------------------------
  
  ! read the individual gridtime files at each timestep and sum
  ! the surface influence over each receptor and the inversion
  ! time interval

  allocate(surfinf(numxgrid,numygrid))

  call calcmeansurfinf(surfinf)

  !---------------------------------------------------------------
  ! read prior fluxes
  !---------------------------------------------------------------

  ! read prior fluxes from ncdf file

  call reademissions()

  allocate(emisflex(numxgrid,numygrid,ntemi))
  allocate(memisflex(numxgrid,numygrid))

  ! map fluxes to flexpart grid (for use in mapgrid)
  ! second argument is unit of output (assumes input unit is kg/m2/h):
  ! unit = 1: area flux (no unit change)
  ! unit = 2: mass (no time unit change)
  ! unit = 3: area flux (change to per second)

  call convertgrid(.true.,3,nxemi,nyemi,lonemi,latemi,ntemi,emis,emisflex)

  ! annual mean emission for use in mapgrid

  memisflex=sum(emisflex,dim=3)

  !---------------------------------------------------------------
  ! calculate inversion grid
  !---------------------------------------------------------------

  ! calculate the grid for inversion by aggregating grid cells of 
  ! low surface influence together

  ! alternate inversion grid definitions can be used for
  ! nbox_xy(numxnest,numynest) in which case replace
  ! routine mapgrid with a new grid definition

  allocate(xlanduse(numxnest,numynest,nclass))
  allocate(surfinfnest(numxnest,numynest))
  allocate(memisflexnest(numxnest,numynest))
  allocate(nbox_xy(numxnest,numynest))

  call readlanduse()

  ! create grid for inversion (for nested domain)

  surfinfnest=surfinf(ixl:ixr,iyl:iyu)
  memisflexnest=memisflex(ixl:ixr,iyl:iyu)

  if(.not.restart) then
    call mapgrid(surfinfnest,memisflexnest)
  else
    call readnboxxy()
  endif
                
  deallocate(surfinfnest)
  deallocate(memisflexnest)

  ! read land-sea mask

  call readlsm()

  ! convert land-sea mask to flexpart grid

  allocate(lsmflex(numxgrid,numygrid,1))

  call convertgrid(.false.,1,nxlsm,nylsm,lonlsm,latlsm,1,lsm,lsmflex)

  ! read ice-snow mask
  
  call readice()

  ! convert ice-snow mask to flexpart grid

  allocate(iceflex(numxgrid,numygrid,1))

  call convertgrid(.false.,1,nxice,nyice,lonice,latice,1,ice,iceflex) 

  ! calculate area of boxes on inversion grid

  call boxarea() 

  ! calculate grid interpolation operator based on nbox_xy

  call gridoperator()

  ! calculate grid interpolation operator for coarse grid used
  ! for calculating the background using a global model

  if(calcbg) then
    call gridcoarse()
    if(optimizbg) then
      nbkgnd=ncrs
    else
      nbkgnd=0
    endif
  else
   ncrs=1
   nbkgnd=0
  endif

  !---------------------------------------------------------------
  ! calculate the state vector 
  !---------------------------------------------------------------

  ! scale fluxes and flux error for better numerical handling

  emisflex=emisflex*numscale
  maxerr=maxerr*numscale
  minerr=minerr*numscale

  ! map fluxes to inversion grid and vectorize

  allocate(emisprior(numxgrid,numygrid,nemsteps))
  allocate(xb(nbox*nemsteps+nbkgnd))
  allocate(timeout(nemsteps))
  xb(:)=0.

  call calcstatevector(xb,timeout)

  !---------------------------------------------------------------
  ! read observations
  !---------------------------------------------------------------

  if(.not.restart) then
    call readobs()
  endif
 
  !---------------------------------------------------------------
  ! calculate the source-receptor matrix (kernel)
  !---------------------------------------------------------------

  ! reads surface influences for each timestep and receptor and
  ! writes these to the source-receptor matrix for the nested 
  ! domain (knest) and for outside the nested domain (kout)

  allocate(knest(maxrows,nemsteps*nbox))
  allocate(kout(maxrows,ncrs))
  kout(:,:)=0.
  knest(:,:)=0.

  if(.not.lowmemory) then
    allocate(knest_fin(maxrows,numxnest*numynest))
    knest_fin(:,:)=0.
  endif
 
  ! if first time run, generate source-receptor matrix and save it
  ! otherwise read source-receptor matrix from file

  if(.not.restart) then
    call calcsrmatrix(knest,kout)
    call savetransport(knest,kout)
  else
    call readtransport(knest,kout)
  endif

  ! scale knest and knest_fin for better numerical handling

  knest=knest/numscale

  if(.not.lowmemory) then
    knest_fin=knest_fin/numscale
  endif

  !---------------------------------------------------------------
  ! calculate background mixing ratios
  !---------------------------------------------------------------

  allocate(bkgnd(maxrows,ncrs))
  bkgnd(:,:)=0.

  if(calcbg) then
    call background(bkgnd)
  endif

  !---------------------------------------------------------------
  ! average observations and kernel to match
  !---------------------------------------------------------------

  ! average observations and kernel to obstep 
  ! note: observations from moving receptors are not averaged

  allocate(allobs(maxobs,6))
  allocate(knest_obs(maxobs,nemsteps*nbox))
  allocate(kout_obs(maxobs,ncrs))
  allocate(bkgnd_obs(maxobs,ncrs))
  allobs(:,:)=-999.
  knest_obs(:,:)=0.
  kout_obs(:,:)=0.
  bkgnd_obs(:,:)=0.

  if(.not.lowmemory) then
    allocate(knest_finobs(maxobs,numxnest*numynest))
    knest_finobs(:,:)=0.
  endif

  call averageobs(knest,kout,bkgnd,knest_obs,kout_obs,bkgnd_obs,allobs)

  !---------------------------------------------------------------
  ! calculate prior mixing ratios
  !---------------------------------------------------------------

  allocate(yobs(maxobs))
  allocate(ymod(maxobs))
  allocate(ybgi(maxobs))
  allocate(yunc(maxobs))
  allocate(ydat(maxobs))
  allocate(yrec(maxobs))

  deallocate(knest)
  allocate(knest_trim(maxobs,nemsteps*nbox+nbkgnd))

  if(.not.lowmemory) then
    deallocate(knest_fin)
    allocate(knest_fintrim(maxobs,numxnest*numynest))
  endif

  call calcprior(knest_obs,knest_trim,kout_obs,&
                 &bkgnd_obs,xb,allobs,yobs,ymod,ybgi,yunc,ydat,yrec,nobs)

  deallocate(knest_obs)
  deallocate(kout_obs)

  if(.not.lowmemory) then
    deallocate(knest_finobs)
  endif
  
  ! if checkprior = true: save output and exit
  if(checkprior) go to 10

  !---------------------------------------------------------------
  ! calculate prior flux error
  !---------------------------------------------------------------

  if(.not.precalcerror) call calcpriorerror(outheight)
  if(precalcerror) call readpriorerror(outheight)

  !---------------------------------------------------------------
  ! calculate prior error covariance matrix
  !---------------------------------------------------------------

  call errorcovariance_coarse()

  !---------------------------------------------------------------
  ! calculate observation error covariance matrix
  !---------------------------------------------------------------

  allocate(covr(nobs,nobs))

  call obscovariance(covr,yunc,nobs)

  !---------------------------------------------------------------
  ! calculate spatial aggregation error
  !---------------------------------------------------------------

  allocate(covagg(nobs,nobs))

  if(.not.lowmemory) then
    ! calculate aggregation error
    call erroraggreg(numxnest*numynest,maxobs,nobs,nbox,covagg)  
    ! add aggregation error covariance matrix to observation error covariance matrix
    covr=covr+real(covagg,kind=4) 
  endif

  !---------------------------------------------------------------
  ! add random perturbation to xb and yobs
  !---------------------------------------------------------------

  ! in this case xb are the 'true' emissions and ymod the observations
  ! xb are perturbed to form the prior and yobs are perturbed to introduce noise

  if(perturb) then
    ! returns xb and yobs with perturbations
    call calcperturb(nbox*nemsteps+nbkgnd,maxobs,nobs,covr,xb,yobs,ymod)
  endif

  !---------------------------------------------------------------
  ! calculate cost a priori
  !---------------------------------------------------------------
  
  call calccost(1,nbox*nemsteps+nbkgnd,maxobs,nobs,xb,xb,yobs,knest_trim,covr,cost)

  !---------------------------------------------------------------
  ! inversion calculations
  !---------------------------------------------------------------

  allocate(gain(nbox*nemsteps+nbkgnd,nobs))
  allocate(cova(nemsteps*nbox+nbkgnd,nemsteps*nbox+nbkgnd))
  allocate(x(nbox*nemsteps+nbkgnd))

  ! fewer observations than state variables

  if(nobs.le.nemsteps*nbox+nbkgnd) then
    call bayesian1(nbox*nemsteps+nbkgnd,maxobs,nobs,covr,yobs,knest_trim,xb,x,cova,gain)
  endif

  ! fewer state variables than observations

  if(nobs.gt.nemsteps*nbox+nbkgnd) then
    call bayesian2(nbox*nemsteps+nbkgnd,maxobs,nobs,covr,yobs,knest_trim,xb,x,cova,gain)
  endif

  !---------------------------------------------------------------
  ! information content
  !---------------------------------------------------------------

  call infocontent(nbox*nemsteps+nbkgnd,maxobs,nobs,knest_trim,gain)

  !---------------------------------------------------------------
  ! calculate cost a posteriori
  !---------------------------------------------------------------

  call calccost(2,nbox*nemsteps+nbkgnd,maxobs,nobs,x,xb,yobs,knest_trim,covr,cost)

  !---------------------------------------------------------------
  ! map fluxes and errors back to flexpart grid
  !---------------------------------------------------------------

  allocate(errb(nbox*nemsteps))
  allocate(erra(nbox*nemsteps))
  allocate(emispost(numxnest,numynest,nemsteps))
  allocate(errpost(numxnest,numynest,nemsteps))
  allocate(errpri(numxnest,numynest,nemsteps))

  ! posterior flux errors
  do n=1,nbox*nemsteps
    erra(n)=real(sqrt(abs(sum(cova(:,n)))),kind=4)
  end do

  ! prior flux errors
  ! use eigendecomposition of B
  allocate(tmp(1,neig))
  allocate(tmp1(nbox,1))
  errb(:)=0.
  do jt=1,nemsteps
    do k=1,nbox
      do i=1,nbox
        ! calculate column vector of U*D*U^T
        do j=1,neig
          ! calculate row vector of U*D
          tmp(1:1,j:j)=bevec(i,j)*beval(j)
        end do
        tmp1(i:i,1:1)=matmul(tmp,reshape(bevec(k,:),(/neig,1/)))
      end do
      do it=1,nemsteps
        errb((jt-1)*nbox+k)=errb((jt-1)*nbox+k)+cortime(jt,it)*real(sum(tmp1),kind=4)
      end do
    end do ! k
  end do ! jt
  ! take sqrt of variances
  do n=1,nbox*nemsteps
    errb(n)=sqrt(errb(n))
  end do

  ! if optimize background: prior and posterior background scalar errors
  allocate(ebgb(nbkgnd))
  allocate(ebga(nbkgnd))
  if(optimizbg) then
    do n=1,nbkgnd
      ebgb(n)=0.01 ! assume 1% errors
      ebga(n)=real(sqrt(cova(nbox*nemsteps+n,nbox*nemsteps+n)),kind=4)
    end do
  endif

  ! re-grid posterior emissions and errors on fine grid

  if(.not.lowmemory.and.scaleflx.eq.2) then
    ! use Bayesian optimization 
    call regridoptimize(numxnest*numynest,nbox,cova,xb,x)
  else
    ! use prior information
    call regridemissions(x,xb,emispost)
    call regriderrors(erra,errpost)
    call regriderrors(errb,errpri)
  endif

  !---------------------------------------------------------------
  ! calculate prior and posterior concentrations
  !---------------------------------------------------------------

  ! re-scale fluxes and errors 
  emisflex=emisflex/numscale
  emisprior=emisprior/numscale
  emispost=emispost/numscale
  errpost=errpost/numscale
  errpri=errpri/numscale

  ! read gridded flexpart output files to calculate surface influence
  ! and multiply with the prior and posterior fluxes

  allocate(modout(maxrows,8))

  write(logid,*) 'Calculating posterior model mixing ratios at full resolution...'

  call timeseries(bkgnd,kout,timeout,modout,nmod,xb,x)

  !---------------------------------------------------------------
  ! save output
  !---------------------------------------------------------------

  call save(nobs,knest_trim,covr,covagg,&
            &cova,gain,xb,x,yrec,ydat,yobs,ymod,ybgi,&
            &yunc,ebgb,ebga,modout,nmod,timeout)

10 continue

  if(checkprior) then
    emisflex=emisflex/numscale
    call save_forward(nobs,yrec,ydat,yobs,ymod,ybgi,yunc)
  endif

  !---------------------------------------------------------------
  ! prepare to exit
  !---------------------------------------------------------------

  deallocate(outheight)
  deallocate(releases)
  deallocate(surfinf)
  deallocate(lonemi)
  deallocate(latemi)
  deallocate(timemi)
  deallocate(emis)
  deallocate(emisflex)
  deallocate(memisflex)
  deallocate(xlanduse)
  deallocate(nbox_xy)
  deallocate(lonlsm)
  deallocate(latlsm)
  deallocate(lsm)
  deallocate(lsmflex)
  deallocate(gridoper)
  deallocate(indexobs)
  deallocate(knest_trim)
  if(.not.lowmemory) then
    deallocate(knest_fintrim)
  endif
  deallocate(kout)
  deallocate(allobs)
  deallocate(bkgnd)
  deallocate(bkgnd_obs)
  deallocate(emisprior)
  deallocate(xb)
  deallocate(timeout)
  deallocate(yobs)
  deallocate(ymod)
  deallocate(ybgi)
  deallocate(yunc)
  deallocate(ydat)
  deallocate(yrec)

  if(checkprior) go to 20 

  deallocate(bevec)
  deallocate(beval)
  deallocate(covagg)
  if(.not.lowmemory) then
    deallocate(bfinevec)
    deallocate(bfineval)
    if(scaleflx.eq.2) then
      deallocate(bfinawevec)
      deallocate(bfinaweval)
    endif
  endif 
  deallocate(covr)
  deallocate(gain)
  deallocate(cova)
  deallocate(x)
  deallocate(errb)
  deallocate(erra)
  deallocate(emispost)
  deallocate(errpost)
  deallocate(errprior)
  deallocate(modout)
  deallocate(area_grid)
  deallocate(gridlon)
  deallocate(gridlat)
  deallocate(nestlon)
  deallocate(nestlat)
  deallocate(area_box)
  deallocate(dimx)
  deallocate(dimy)
  deallocate(dimxnest)
  deallocate(dimynest)
  deallocate(dimxynest)
  deallocate(dimxy)
  deallocate(dimr)
  deallocate(dimo)
  deallocate(dimbox)
  deallocate(dimb)
  deallocate(dimbx)
  deallocate(recname)
  deallocate(rectyp)  
 
20 continue

  write(logid,*) 'Good-bye'
  close(logid)

end program main

