!---------------------------------------------------------------------------------------
! FLEXINVERT : readlmdz 
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Reads mixing ratios from global fields from LMDZ model runs (LMDZ96 grid) 
!! note:
!!  - particular set-up is for LMDZ96 output files (runch4_online_test5)
!!  - time in LMDZ96 file is seconds since 1979-01-01 00:00:00
!!  - pressure in LMDZ96 is Pa
!!  - variable in LMDZ96 is volume mixing ratio
!---------------------------------------------------------------------------------------

subroutine readlmdz(jd)

  use mod_var
  use mod_dates

  implicit none

  include 'netcdf.inc'

  character(len=100) :: filename,filepart1,filepart2,filepart3
  character(len=4) :: yyyystr
  character(len=2) :: mmstr
  integer :: i,kz,ix,jy,ind1,ind2,cnt
  integer :: yyyymmddhh,yyyymmdd,yyyymm,dd,mm,yyyy
  integer :: nid,yid,xid,zid,tid,ierr,varid
  real(kind=8) :: jd,jdi,jdf,jdt
  real :: eomday,tmp,prs1,prs2
  real, dimension(:), allocatable :: lato
  integer, dimension(:), allocatable :: indy
  real, dimension(:), allocatable :: ap,bp
  real, dimension(:,:), allocatable :: help
  real, dimension(:,:,:), allocatable :: psurf
  real, dimension(:,:,:,:), allocatable :: temp,immrtmp


  ! calculate start date
  if((jd-trajlen).ge.juldatei) then
    jdi=jd-trajlen
  else
    jdi=jd
  endif
  yyyymmddhh=caldate(jdi)
  yyyymm=yyyymmddhh/10000
  jdi=juldate(yyyymm*100+1,0)

  ! calculate end date 
  yyyymmddhh=caldate(jd)
  yyyymm=yyyymmddhh/10000
  eomday=calceomday(yyyymm)
  jdf=jd+eomday-1
  yyyymmddhh=caldate(jdf)

  nday=int(jdf-jdi)+1

  ! initialize 
  jdt=jdi
  i=0
  cnt=0

  ! loop over monthly files

  do while(jdt.le.jdf)

    i=i+1

    ! define exact filename
    yyyymmddhh=caldate(jdt)
    yyyymmdd=yyyymmddhh/100
    yyyymm=yyyymmdd/100
    yyyy=yyyymmddhh/1000000
    mm=(yyyymmddhh-yyyy*1000000)/10000
    ind1=index(fileimmr,'YYYY')-1
    filepart1=fileimmr(1:ind1)
    ind2=index(fileimmr,'MM')-1
    filepart2=fileimmr((ind1+5):ind2)
    filepart3=fileimmr((ind2+3):len(fileimmr))
    write(yyyystr,'(i4)') yyyy
    write(mmstr,'(i2.2)') mm
    filename=trim(filepart1)//yyyystr//trim(filepart2)//mmstr//trim(filepart3)

    ! open netcdf file
    write(logid,*) 'Reading the file: '//trim(filename)
    ierr=nf_open(trim(filename),NF_NOWRITE,nid)
    if(ierr.ne.0) then
      write(logid,*) 'Error opening file: '//trim(filename)
      stop
    endif

    ! inquire about variables
    ierr=nf_inq_dimid(nid,'x',xid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: x'
      stop
    endif
    ierr=nf_inq_dimid(nid,'y',yid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: y'
      stop
    endif
    ierr=nf_inq_dimid(nid,'presnivs',zid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: presnivs'
      stop
    endif
    ierr=nf_inq_dimid(nid,'time_counter',tid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: time_counter'
      stop
    endif

    ! read dimension sizes
    ierr=nf_inq_dimlen(nid,xid,nxini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,yid,nyini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,zid,nzini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,tid,ntini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif

    ! allocate variables
    if(i.eq.1) then
      ! allocate only on first iteration
      allocate(ap(20))
      allocate(bp(20))
      allocate(lonini(nxini))
      allocate(latini(nyini))
      allocate(help(nxini,nyini))
      allocate(lato(nyini))
      allocate(indy(nyini))
      allocate(psurf(nxini,nyini,ntini))
      allocate(temp(nxini,nyini,nzini,ntini))
      allocate(immrraw(nxini,nyini,nzini,nday))
    endif
    ! allocate each iteration
    allocate(immrtmp(nxini,nyini,nzini,ntini))

    ! read variables
    ! read only on first iteration
    if(i.eq.1) then
      ierr=nf_inq_varid(nid,'ap',varid)
      ierr=nf_get_var_real(nid,varid,ap)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: ap'
        stop
      endif
      ierr=nf_inq_varid(nid,'bp',varid)
      ierr=nf_get_var_real(nid,varid,bp)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: bp'
        stop
      endif
      ierr=nf_inq_varid(nid,'nav_lon',varid)
      ierr=nf_get_var_real(nid,varid,help)
      lonini=help(:,1)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: nav_lon'
        stop
      endif
      ierr=nf_inq_varid(nid,'nav_lat',varid)
      ierr=nf_get_var_real(nid,varid,help)
      latini=help(1,:)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: nav_lat'
        stop
      endif
      ierr=nf_inq_varid(nid,'ps',varid)
      ierr=nf_get_var_real(nid,varid,psurf)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: presnivs'
        stop
      endif
      ierr=nf_inq_varid(nid,'temp',varid)
      ierr=nf_get_var_real(nid,varid,temp)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: temp'
        stop
      endif
    endif
    ! read on each iteration
    ierr=nf_inq_varid(nid,trim(bginame),varid)
    ierr=nf_get_var_real(nid,varid,immrtmp)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading variable: ',trim(bginame)
      stop
    endif

    ! close file
    ierr=nf_close(nid)

    ! assign to immrraw 
    immrraw(:,:,:,(cnt+1):(cnt+ntini))=immrtmp
    deallocate(immrtmp)

    eomday=calceomday(yyyymm)
    jdt=jdt+eomday
    cnt=cnt+ntini

  enddo 

  ! if latitude order N-S convert to S-N
  lato=latini
  call sort(nyini,lato)
  ! indices to lat
  do jy=1,nyini
    indy(jy)=minloc(lato,dim=1,mask=lato.eq.latini(jy))
  end do
  latini=lato

  ! calculate geopotential height for each gridcell
  allocate(geoh(nxini,nyini,nzini))
  do kz=1,nzini
    do jy=1,nyini
      do ix=1,nxini
        prs1=sum(psurf(ix,jy,:))/real(size(psurf,3))
        prs2=ap(kz)+bp(kz)*sum(psurf(ix,jy,:))/real(size(psurf,3))
        tmp=sum(temp(ix,jy,kz,:))/real(size(temp,4))
        geoh(ix,jy,kz)=(gasc*tmp/gc/mmol)*log(prs1/prs2)
      enddo
    enddo
  enddo  

  ! reverse latitude order from N-S to S-N
  immrraw=immrraw(:,indy,:,:)
  geoh=geoh(:,indy,:)

  ! calculate timestep as decimal year
  allocate(timini(nday))
  do i=1,nday
    yyyymmddhh=caldate(jdi+i-1)
    yyyy=yyyymmddhh/1000000
    mm=(yyyymmddhh-1000000*yyyy)/10000
    dd=(yyyymmddhh-1000000*yyyy-10000*mm)/100
    timini(i)=float(yyyy)+float(mm-1)/12.+float(dd-1)/365.
  end do

  ! prepare to exit
  deallocate(ap)
  deallocate(bp)
  deallocate(lato)
  deallocate(indy)
  deallocate(psurf)
  deallocate(temp)  
  deallocate(help)

end subroutine readlmdz

