!---------------------------------------------------------------------------------------
! FLEXINVERT : readpyvar
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Reads mixing ratios from global fields from PYVAR model runs (LMDZ96 grid) 
!! note:
!!  - particular set-up is for PYVAR output files (qfield files)
!!  - time is hours since the start of each month
!!  - pressure is calculated from LMDZ coordinates (19 levels)
!!  - variable unit is volume mixing ratio
!---------------------------------------------------------------------------------------

subroutine readpyvar(jd)
 
  use mod_var
  use mod_dates

  implicit none

  include 'netcdf.inc'

  character(len=100) :: filename,filepart1,filepart2,filepart3
  character(len=4) :: yyyystr
  character(len=2) :: mmstr
  integer :: i,kz,ix,jy,ind,ind1,ind2,cnt
  integer :: yyyymmddhh,yyyymmdd,yyyymm,yyyy,mm,dd,hh
  integer :: nid,yid,xid,zid,tid,ierr,varid
  real(kind=8) :: jd,jdi,jdf,jdt
  real, dimension(:), allocatable :: lato,pres,temp
  integer, dimension(:), allocatable :: indy
  real, dimension(:,:,:,:), allocatable :: immrtmp
  real :: eomday
  real, dimension(20) :: plev
  real, dimension(1000) :: pz,tz

  ! LMDZ/PYVAR pressure levels
  real, dimension(20) :: ap = (/5.55111512312578e-12,593.108831876127,1399.18307002416,&
    2621.24871786529,4455.05592361762,7065.66628497033,10523.7283606666,&
    14667.8361044022,18887.842576887,21967.0207025829,22444.1964082259,&
    19931.4918336165,15945.0159392323,12032.002896074,8609.5823366703,&
    5772.52298204432,3550.95334329671,1914.74920954818,776.486715808502,0./)
  real, dimension(20) :: bp = (/1.,0.976412024687698,0.944760805608567,0.897554242776951,&
    0.828154015986934,0.731510415984927,0.605607987378476,&
    0.454277335311256,0.291469875868877,0.144243467803651,&
    0.0452531724980914,0.00606048279758092,0.000147155133957582,&
    8.60044395379218e-08,6.12220131697327e-15,0.,0.,0.,0.,0./)


  ! calculate start date
  if((jd-trajlen).ge.juldatei) then
    jdi=jd-trajlen
  else
    jdi=jd
  endif
  yyyymmddhh=caldate(jdi)
  yyyymm=yyyymmddhh/10000
  jdi=juldate(yyyymm*100+1,0)

  ! calculate end date 
  yyyymmddhh=caldate(jd)
  yyyymm=yyyymmddhh/10000
  eomday=calceomday(yyyymm)
  jdf=jd+eomday-1

  nday=int(jdf-jdi)+1

  ! initialize 
  jdt=jdi
  i=0
  cnt=0

  ! loop over monthly files

  do while(jdt.le.jdf)

    i=i+1

    ! define exact filename
    yyyymmddhh=caldate(jdt)
    yyyymmdd=yyyymmddhh/100
    yyyymm=yyyymmdd/100
    yyyy=yyyymmddhh/1000000
    mm=(yyyymmddhh-yyyy*1000000)/10000
    ind1=index(fileimmr,'YYYY')-1
    filepart1=fileimmr(1:ind1)
    ind2=index(fileimmr,'MM')-1
    filepart2=fileimmr((ind1+5):ind2)
    filepart3=fileimmr((ind2+3):len(fileimmr))
    write(yyyystr,'(i4)') yyyy
    write(mmstr,'(i2.2)') mm
    filename=trim(filepart1)//yyyystr//trim(filepart2)//mmstr//trim(filepart3)
  
    ! open netcdf file
    write(logid,*) 'Reading the file: ',trim(filename)
    ierr=nf_open(trim(filename),NF_NOWRITE,nid)
    if(ierr.ne.0) then
      write(logid,*) 'Error opening file: ',trim(filename)
      stop
    endif

    ! inquire about variables
    ierr=nf_inq_dimid(nid,'rlonv',xid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: rlonv'
      stop
    endif
    ierr=nf_inq_dimid(nid,'rlatu',yid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: rlatu'
      stop
    endif
    ierr=nf_inq_dimid(nid,'sigs',zid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: sigs'
      stop
    endif
    ierr=nf_inq_dimid(nid,'time',tid)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension: time'
      stop
    endif

    ! read dimension sizes
    ierr=nf_inq_dimlen(nid,xid,nxini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,yid,nyini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,zid,nzini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif
    ierr=nf_inq_dimlen(nid,tid,ntini)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading dimension length'
      stop
    endif

    ! allocate variables
    if(i.eq.1) then
      ! allocate only on first iteration
      allocate(lonini(nxini))
      allocate(latini(nyini))
      allocate(lato(nyini))
      allocate(indy(nyini))
      allocate(pres(nzini))
      allocate(temp(nzini))
      allocate(immrraw(nxini,nyini,nzini,nday))
    endif
    ! allocate each iteration
    allocate(immrtmp(nxini,nyini,nzini,ntini))

    ! read variables
    if(i.eq.1) then
      ! read only on first iteration
      ierr=nf_inq_varid(nid,'rlonv',varid)
      ierr=nf_get_var_real(nid,varid,lonini)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: rlonv'
        stop
      endif
      ierr=nf_inq_varid(nid,'rlatu',varid)
      ierr=nf_get_var_real(nid,varid,latini)
      if(ierr.ne.0) then
        write(logid,*) 'Error reading variable: rlatu'
        stop
      endif
    endif
    ! read each iteration
    ierr=nf_inq_varid(nid,trim(bginame),varid)
    ierr=nf_get_var_real(nid,varid,immrtmp)
    if(ierr.ne.0) then
      write(logid,*) 'Error reading variable: ',trim(bginame)
      stop
    endif

    ! close file
    ierr=nf_close(nid)

    ! assign to immrraw
    immrraw(:,:,:,(cnt+1):(cnt+ntini))=immrtmp
    deallocate(immrtmp)

    eomday=calceomday(yyyymm)
    jdt=jdt+eomday
    cnt=cnt+ntini

  enddo 

  ! convert lat and lon from radians to degrees
  latini=latini*180./pi
  lonini=lonini*180./pi

  ! convert from mass mixing ratio to volume mixing ratio
  ! for consistency in initial.f90
  ! (in PYVAR files N2O is mass of N = 14 and not molarmass 2N = 28)
  immrraw=immrraw*28.97/14.

  ! calculate pressure levels
  do i=1,nzini+1
    plev(i)=ap(i)+bp(i)*ps
  enddo
  do i=1,nzini
    pres(i)=0.5*(plev(i)+plev(i+1))
  enddo  

  ! calculate temperature levels
  tz(1)=ts
  do i=2,1000
    tz(i)=ts-lrate*float((i-1)*50)
    pz(i)=ps*exp(-mmol*gc*float((i-1)*50)/(gasc*tz(i)))
  enddo    
  ! find temperature to match PYVAR pressure levels
  do i=1,nzini
    ind=minloc(abs(pz(:)-pres(i)),dim=1,mask=abs(pz(:)-pres(i)).eq.minval(abs(pz(:)-pres(i))))
    temp(i)=tz(ind)
  enddo

  ! if latitude order N-S convert to S-N
  lato=latini
  call sort(nyini,lato)
  ! indices to lat
  do jy=1,nyini
    indy(jy)=minloc(lato,dim=1,mask=lato.eq.latini(jy))
  end do
  latini=lato

  ! calculate geopotential height for each gridcell
  allocate(geoh(nxini,nyini,nzini))
  do kz=1,nzini
    do jy=1,nyini
      do ix=1,nxini
        geoh(ix,jy,kz)=(gasc*temp(kz)/gc/mmol)*log(ps/pres(kz))
      enddo
    enddo
  enddo

  ! reverse latitude order from N-S to S-N
  immrraw=immrraw(:,indy,:,:)
  geoh=geoh(:,indy,:)

  ! calculate timestep as decimal year
  allocate(timini(nday))
  do i=1,nday
    yyyymmddhh=caldate(jdi+i-1)
    yyyy=yyyymmddhh/1000000
    mm=(yyyymmddhh-1000000*yyyy)/10000
    dd=(yyyymmddhh-1000000*yyyy-10000*mm)/100
    timini(i)=float(yyyy)+float(mm-1)/12.+float(dd-1)/365.
  end do

  ! prepare to exit
  deallocate(lato)
  deallocate(indy)
  deallocate(pres)
  deallocate(temp) 

end subroutine readpyvar
